__all__ = ("col_map_dict",)


def col_map_dict(dict_name=None):
    if dict_name is None:
        dict_name = "FBS"
    dict_name = dict_name.lower()

    if dict_name == "fbs" or dict_name == "opsimfbs":
        col_map = {}
        col_map["ra"] = "fieldRA"
        col_map["dec"] = "fieldDec"
        col_map["raDecDeg"] = True
        col_map["mjd"] = "observationStartMJD"
        col_map["exptime"] = "visitExposureTime"
        col_map["visittime"] = "visitTime"
        col_map["alt"] = "altitude"
        col_map["az"] = "azimuth"
        col_map["lst"] = "observationStartLST"
        col_map["filter"] = "filter"
        col_map["fiveSigmaDepth"] = "fiveSigmaDepth"
        col_map["night"] = "night"
        col_map["slewtime"] = "slewTime"
        col_map["slewdist"] = "slewDistance"
        col_map["seeingEff"] = "seeingFwhmEff"
        col_map["seeingGeom"] = "seeingFwhmGeom"
        col_map["skyBrightness"] = "skyBrightness"
        col_map["moonDistance"] = "moonDistance"
        col_map["fieldId"] = "fieldId"
        col_map["proposalId"] = "proposalId"
        col_map["slewactivities"] = {}
        col_map["metadataList"] = [
            "airmass",
            "normairmass",
            "seeingEff",
            "skyBrightness",
            "fiveSigmaDepth",
            "HA",
            "moonDistance",
            "solarElong",
            "saturation_mag",
        ]
        col_map["metadataAngleList"] = ["rotSkyPos"]
        col_map["note"] = "note"

    elif dict_name == "opsimv4":
        col_map = {}
        col_map["ra"] = "fieldRA"
        col_map["dec"] = "fieldDec"
        col_map["raDecDeg"] = True
        col_map["mjd"] = "observationStartMJD"
        col_map["exptime"] = "visitExposureTime"
        col_map["visittime"] = "visitTime"
        col_map["alt"] = "altitude"
        col_map["az"] = "azimuth"
        col_map["lst"] = "observationStartLST"
        col_map["filter"] = "filter"
        col_map["fiveSigmaDepth"] = "fiveSigmaDepth"
        col_map["night"] = "night"
        col_map["slewtime"] = "slewTime"
        col_map["slewdist"] = "slewDistance"
        col_map["seeingEff"] = "seeingFwhmEff"
        col_map["seeingGeom"] = "seeingFwhmGeom"
        col_map["skyBrightness"] = "skyBrightness"
        col_map["moonDistance"] = "moonDistance"
        col_map["fieldId"] = "fieldId"
        col_map["proposalId"] = "proposalId"
        # slew speeds table
        col_map["slewSpeedsTable"] = "SlewMaxSpeeds"
        # slew states table
        col_map["slewStatesTable"] = "SlewFinalState"
        # slew activities list
        col_map["slewActivitiesTable"] = "SlewActivities"
        # Slew columns
        col_map["Dome Alt Speed"] = "domeAltSpeed"
        col_map["Dome Az Speed"] = "domeAzSpeed"
        col_map["Tel Alt Speed"] = "telAltSpeed"
        col_map["Tel Az Speed"] = "telAzSpeed"
        col_map["Rotator Speed"] = "rotatorSpeed"
        col_map["Tel Alt"] = "telAlt"
        col_map["Tel Az"] = "telAz"
        col_map["Rot Tel Pos"] = "rotTelPos"
        col_map["Dome Alt"] = "domeAlt"
        col_map["Dome Az"] = "domeAz"
        col_map["slewactivities"] = {
            "Dome Alt": "domalt",
            "Dome Az": "domaz",
            "Dome Settle": "domazsettle",
            "Tel Alt": "telalt",
            "Tel Az": "telaz",
            "Tel Rot": "telrot",
            "Tel Settle": "telsettle",
            "TelOptics CL": "telopticsclosedloop",
            "TelOptics OL": "telopticsopenloop",
            "Readout": "readout",
            "Filter": "filter",
        }
        col_map["metadataList"] = [
            "airmass",
            "normairmass",
            "seeingEff",
            "skyBrightness",
            "fiveSigmaDepth",
            "HA",
            "moonDistance",
            "solarElong",
        ]
        col_map["metadataAngleList"] = ["rotSkyPos"]

    elif dict_name == "opsimv3":
        col_map = {}
        col_map["ra"] = "fieldRA"
        col_map["dec"] = "fieldDec"
        col_map["raDecDeg"] = False
        col_map["mjd"] = "expMJD"
        col_map["exptime"] = "visitExpTime"
        col_map["visittime"] = "visitTime"
        col_map["alt"] = "altitude"
        col_map["az"] = "azimuth"
        col_map["lst"] = "lst"
        col_map["filter"] = "filter"
        col_map["fiveSigmaDepth"] = "fiveSigmaDepth"
        col_map["night"] = "night"
        col_map["slewtime"] = "slewTime"
        col_map["slewdist"] = "slewDist"
        col_map["seeingEff"] = "FWHMeff"
        col_map["seeingGeom"] = "FWHMgeom"
        col_map["skyBrightness"] = "filtSkyBrightness"
        col_map["moonDistance"] = "dist2Moon"
        col_map["fieldId"] = "fieldID"
        col_map["proposalId"] = "propID"
        # slew speeds table
        col_map["slewSpeedsTable"] = "SlewMaxSpeeds"
        # slew states table
        col_map["slewStatesTable"] = "SlewStates"
        # Slew activities list
        col_map["slewActivitiesTable"] = "SlewActivities"
        col_map["Dome Alt Speed"] = "domeAltSpeed"
        col_map["Dome Az Speed"] = "domeAzSpeed"
        col_map["Tel Alt Speed"] = "telAltSpeed"
        col_map["Tel Az Speed"] = "telAzSpeed"
        col_map["Rotator Speed"] = "rotatorSpeed"
        col_map["Tel Alt"] = "telAlt"
        col_map["Tel Az"] = "telAz"
        col_map["Rot Tel Pos"] = "rotTelPos"
        col_map["Dome Alt"] = "domAlt"
        col_map["Dome Az"] = "domAz"
        col_map["slewactivities"] = {
            "Dome Alt": "DomAlt",
            "Dome Az": "DomAz",
            "Tel Alt": "TelAlt",
            "Tel Az": "TelAz",
            "Tel Rot": "Rotator",
            "Settle": "Settle",
            "TelOptics CL": "TelOpticsCL",
            "TelOptics OL": "TelOpticsOL",
            "Readout": "Readout",
            "Filter": "Filter",
        }
        col_map["metadataList"] = [
            "airmass",
            "normairmass",
            "seeingEff",
            "skyBrightness",
            "fiveSigmaDepth",
            "HA",
            "moonDistance",
            "solarElong",
        ]
        col_map["metadataAngleList"] = ["rotSkyPos"]

    elif dict_name == "barebones":
        col_map = {}
        col_map["ra"] = "ra"
        col_map["dec"] = "dec"
        col_map["raDecDeg"] = True
        col_map["mjd"] = "mjd"
        col_map["exptime"] = "exptime"
        col_map["visittime"] = "exptime"
        col_map["alt"] = "alt"
        col_map["az"] = "az"
        col_map["filter"] = "filter"
        col_map["fiveSigmaDepth"] = "fivesigmadepth"
        col_map["night"] = "night"
        col_map["slewtime"] = "slewtime"
        col_map["slewdist"] = None
        col_map["seeingGeom"] = "seeing"
        col_map["seeingEff"] = "seeing"
        col_map["metadataList"] = [
            "airmass",
            "normairmass",
            "seeingEff",
            "skyBrightness",
            "fiveSigmaDepth",
            "HA",
        ]
        col_map["metadataAngleList"] = ["rotSkyPos"]

    else:
        raise ValueError(f"No built in column dict with name {dict_name}")

    return col_map



__all__ = ("CoaddStacker",)

import numpy as np
import numpy.lib.recfunctions as rf

from rubin_sim.maf.stackers import BaseStacker


class CoaddStacker(BaseStacker):
    """
    Stacker to estimate m5 "coadded" per band and par night

    Parameters
    ----------
    mjd_col : `str`, optional
        Name of the MJD column
    ra_col : `str`, optional
        Name of the RA column
    dec_col : `str`, optional
        Name of the Dec column
    m5_col : `str`, optional
        Name of the m5 column
    filter_col : `str`, optional
        Name of the filter column
    night_col : `str`, optional
        Name of the night column
    num_exposures_col : `str`, optional
        Name of the number of exposures per visit column
    visit_time_col : `str`, optional
        Name of the total visit time column
    visit_exposure_time_col : `str`, optional
        Name of the on-sky exposure time column
    """

    cols_added = ["coadd"]

    def __init__(
        self,
        mjd_col="observationStartMJD",
        ra_col="fieldRA",
        dec_col="fieldDec",
        m5_col="fiveSigmaDepth",
        filter_col="filter",
        night_col="night",
        num_exposures_col="numExposures",
        visit_time_col="visitTime",
        visit_exposure_time_col="visitExposureTime",
    ):
        self.cols_req = [
            mjd_col,
            ra_col,
            dec_col,
            m5_col,
            filter_col,
            night_col,
            num_exposures_col,
            visit_time_col,
            visit_exposure_time_col,
        ]
        self.ra_col = ra_col
        self.dec_col = dec_col
        self.night_col = night_col
        self.filter_col = filter_col
        self.m5_col = m5_col
        self.num_exposures_col = num_exposures_col
        self.visit_time_col = visit_time_col
        self.visit_exposure_time_col = visit_exposure_time_col

        self.units = ["int"]

    def _run(self, sim_data, cols_present=False):
        """
        Parameters
        ------------
        sim_data : simulation data
        cols_present: to check whether the field has already been estimated

        Returns
        --------
        numpy array of initial fields plus modified fields:
        - m5Col: "coadded" m5
        - numExposuresCol: sum of  numExposuresCol
        - visitTimeCol: sum of visitTimeCol
        - visitExposureTimeCol: sum of visitExposureTimeCol
        - all other input fields except band (Ra, Dec, night) : median(field)

        """

        if cols_present:
            # Column already present in data; assume it is correct and does not need recalculating.
            return sim_data
        self.dtype = sim_data.dtype
        r = []
        for ra, dec, band in np.unique(sim_data[[self.ra_col, self.dec_col, self.filter_col]]):
            idx = np.abs(sim_data[self.ra_col] - ra) < 1.0e-5
            idx &= np.abs(sim_data[self.dec_col] - dec) < 1.0e-5
            idx &= sim_data[self.filter_col] == band

            sel = sim_data[idx]
            for night in np.unique(sel[self.night_col]):
                idxb = sel[self.night_col] == night
                r.append(tuple(self.fill(sel[idxb])))

        myarray = np.array(r, dtype=self.dtype)
        return myarray

    def fill(self, tab):
        """
        Estimation of new fields (m5 "coadded" values, ...)

        Parameters
        ---------------
        tab : array of (initial) data


        Returns
        -----------
        tuple with modified field values:
         - m5Col: "coadded" m5
        - numExposuresCol: sum of  numExposuresCol
        - visitTimeCol: sum of visitTimeCol
        - visitExposureTimeCol: sum of visitExposureTimeCol
        - all other input fields except band (Ra, Dec, night) : median(field)
        """

        r = []

        for colname in self.dtype.names:
            if colname not in [
                self.m5_col,
                self.num_exposures_col,
                self.visit_time_col,
                self.visit_exposure_time_col,
                self.filter_col,
            ]:
                if colname == "coadd":
                    r.append(1)
                else:
                    if tab[colname].dtype == "object":
                        r.append(tab[colname][0])
                    else:
                        r.append(np.median(tab[colname]))
            if colname == self.m5_col:
                r.append(self.m5_coadd(tab[self.m5_col]))
            if colname in [
                self.num_exposures_col,
                self.visit_time_col,
                self.visit_exposure_time_col,
            ]:
                r.append(np.sum(tab[colname]))
            if colname == self.filter_col:
                r.append(np.unique(tab[self.filter_col])[0])

        return r

    def m5_coadd(self, m5):
        """
        Estimation of "coadded" m5 values based on:
        flux_5sigma = 10**(-0.4*m5)
        sigmas = flux_5sigma/5.
        sigma_tot = 1./sqrt(np.sum(1/sigmas**2))
        flux_tot = 5.*sigma_tot

        Parameters
        ---------------
        m5 : set of m5 (five-sigma depths) values

        Returns
        -----------
        "coadded" m5 value
        """
        return 1.25 * np.log10(np.sum(10.0 ** (0.8 * m5)))



__all__ = ("data_dict", "rs_download_data", "get_data_dir")

import argparse

from rubin_scheduler.data import DEFAULT_DATA_URL, download_rubin_data
from rubin_scheduler.data import get_data_dir as gdd


def get_data_dir():
    """For backwards compatibility since this got moved over to the scheduler."""
    return gdd()


def data_dict():
    """Creates a `dict` for all data buckets and the tar file they map to.
    To create tar files and follow any sym links, run:
        ``tar -chvzf maf_may_2021.tgz maf``

    Returns
    -------
    result : `dict`
        Data bucket filenames dictionary with keys:
        ``"name"``
            Data bucket name (`str`).
        ``"version"``
            Versioned file name (`str`).
    """
    file_dict = {
        "maf": "maf_2022_08_26.tgz",
        "maps": "maps_2022_2_28.tgz",
        "movingObjects": "movingObjects_oct_2021.tgz",
        "orbits": "orbits_2022_3_1.tgz",
        "orbits_precompute": "orbits_precompute_2023_05_23.tgz",
        "sim_baseline": "sim_baseline_2023_09_22.tgz",
        "skybrightness": "skybrightness_2023_09_11.tgz",
        "throughputs": "throughputs_2023_09_22.tgz",
        "tests": "tests_2022_10_18.tgz",
    }
    return file_dict


def rs_download_data():
    """Download data."""

    files = data_dict()
    parser = argparse.ArgumentParser(description="Download data files for rubin_sim package")
    parser.add_argument(
        "--versions",
        dest="versions",
        default=False,
        action="store_true",
        help="Report expected versions, then quit",
    )
    parser.add_argument(
        "-d",
        "--dirs",
        type=str,
        default=None,
        help="Comma-separated list of directories to download",
    )
    parser.add_argument(
        "-f",
        "--force",
        dest="force",
        default=False,
        action="store_true",
        help="Force re-download of data directory(ies)",
    )
    parser.add_argument(
        "--url_base",
        type=str,
        default=DEFAULT_DATA_URL,
        help="Root URL of download location",
    )
    parser.add_argument(
        "--tdqm_disable",
        dest="tdqm_disable",
        default=False,
        action="store_true",
        help="Turn off tdqm progress bar",
    )
    parser.add_argument(
        "--update",
        dest="update",
        default=False,
        action="store_true",
        help="Update versions of data on disk to match current",
    )

    args = parser.parse_args()

    download_rubin_data(
        files,
        dirs=args.dirs,
        print_versions_only=args.versions,
        force=args.force,
        url_base=args.url_base,
        tdqm_disable=args.tdqm_disable,
        update=args.update,
    )



__all__ = ("DirectObs",)

import datetime
import logging
import warnings

import numpy as np
from rubin_scheduler.utils import angular_separation

from .base_obs import BaseObs


class DirectObs(BaseObs):
    """
    Generate observations of a set of moving objects: exact ephemeris at the times of each observation.

    First generates observations on a rough grid and looks for observations within a specified tolerance
    of the actual observations; for the observations which pass this cut, generates a precise ephemeris
    and checks if the object is within the FOV.

    Parameters
    ----------
    footprint: `str`, optional
        Specify the footprint for the FOV. Options include "camera", "circle", "rectangle".
        'Camera' means use the actual LSST camera footprint (following a rough cut with a circular FOV).
        Default is circular FOV.
    r_fov : `float`, optional
        If footprint is "circular", this is the radius of the fov (in degrees).
        Default 1.75 degrees.
    x_tol : `float`, optional
        If footprint is "rectangular", this is half of the width of the (on-sky) fov in the RA
        direction (in degrees).
        Default 5 degrees. (so size of footprint in degrees will be 10 degrees in the RA direction).
    y_tol : `float`, optional
        If footprint is "rectangular", this is half of the width of the fov in Declination (in degrees).
        Default is 3 degrees (so size of footprint in degrees will be 6 degrees in the Dec direction).
    eph_mode: `str`, optional
        Mode for ephemeris generation - nbody or 2body. Default is nbody.
    prelim_eph_mode: str, optional
        Mode for preliminary ephemeris generation, if any is done. Default is 2body.
    eph_type: `str`, optional
        Type of ephemerides to generate - full or basic.
        Full includes all values calculated by openorb; Basic includes a more basic set.
        Default is Basic.  (this includes enough information for most standard MAF metrics).
    eph_file: `str` or None, optional
        The name of the planetary ephemerides file to use for ephemeris generation.
        Default (None) will use the default for PyOrbEphemerides.
    obs_code: `str`, optional
        Observatory code for ephemeris generation. Default is "I11" - Cerro Pachon.
    obs_time_col: `str`, optional
        Name of the time column in the obsData. Default 'observationStartMJD'.
    obs_time_scale: `str`, optional
        Type of timescale for MJD (TAI or UTC currently). Default TAI.
    seeing_col: `str`, optional
        Name of the seeing column in the obsData. Default 'seeingFwhmGeom'.
        This should be the geometric/physical seeing as it is used for the trailing loss calculation.
    visit_exp_time_col: `str`, optional
        Name of the visit exposure time column in the obsData. Default 'visitExposureTime'.
    obs_ra: `str`, optional
        Name of the RA column in the obsData. Default 'fieldRA'.
    obs_dec: `str`, optional
        Name of the Dec column in the obsData. Default 'fieldDec'.
    obs_rot_sky_pos: `str`, optional
        Name of the Rotator column in the obsData. Default 'rotSkyPos'.
    obs_degrees: `bool`, optional
        Whether the observational data is in degrees or radians. Default True (degrees).
    outfile_name : `str`, optional
        The output file name.
        Default is 'lsst_obs.dat'.
    obs_info : `str`, optional
        A string that captures provenance information about the observations.
        For example: 'baseline_v2.0_10yrs, MJD 59853-61677' or 'baseline2018a minus NES'
        Default ''.
    tstep: `float`, optional
        The time between initial (rough) ephemeris generation points, in days.
        Default 1 day.
    rough_tol: `float`, optional
        The initial rough tolerance value for positions, used as a first cut to identify potential
        observations (in degrees).
        Default 10 degrees.
    pre_comp_tol : float (2.08)
        The radial tolerance to add when using pre-computed orbits. Should be
        larger than the full field of view extent.
    """

    def __init__(
        self,
        footprint="camera",
        r_fov=1.75,
        x_tol=5,
        y_tol=3,
        eph_mode="nbody",
        prelim_eph_mode="2body",
        eph_type="basic",
        obs_code="I11",
        eph_file=None,
        obs_time_col="observationStartMJD",
        obs_time_scale="TAI",
        seeing_col="seeingFwhmGeom",
        visit_exp_time_col="visitExposureTime",
        obs_ra="fieldRA",
        obs_dec="fieldDec",
        obs_rot_sky_pos="rotSkyPos",
        obs_degrees=True,
        outfile_name="lsst_obs.dat",
        obs_info="",
        tstep=1.0,
        rough_tol=10.0,
        verbose=True,
        night_col="night",
        filter_col="filter",
        m5_col="fiveSigmaDepth",
        obs_id_col="observationId",
        pre_comp_tol=2.08,
    ):
        super().__init__(
            footprint=footprint,
            r_fov=r_fov,
            x_tol=x_tol,
            y_tol=y_tol,
            eph_mode=eph_mode,
            eph_type=eph_type,
            obs_code=obs_code,
            eph_file=eph_file,
            obs_time_col=obs_time_col,
            obs_time_scale=obs_time_scale,
            seeing_col=seeing_col,
            visit_exp_time_col=visit_exp_time_col,
            obs_ra=obs_ra,
            obs_dec=obs_dec,
            obs_rot_sky_pos=obs_rot_sky_pos,
            obs_degrees=obs_degrees,
            outfile_name=outfile_name,
            obs_info=obs_info,
        )
        self.verbose = verbose
        self.pre_comp_tol = pre_comp_tol
        self.filter_col = filter_col
        self.night_col = night_col
        self.m5_col = m5_col
        self.obs_id_col = obs_id_col
        self.tstep = tstep
        self.rough_tol = rough_tol
        if prelim_eph_mode not in ("2body", "nbody"):
            raise ValueError("Ephemeris generation must be 2body or nbody.")
        self.prelim_eph_mode = prelim_eph_mode

    def run(self, orbits, obs_data, object_positions=None, object_mjds=None):
        """Find and write the observations of each object to disk.

        For each object, generate a very rough grid of ephemeris points (typically using 2body integration).
        Then identify pointings in obs_data which are within

        Parameters
        ----------
        orbits : `rubin_sim.moving_objects.Orbits`
            The orbits to generate ephemerides for.
        obs_data : `np.ndarray`
            The simulated pointing history data.
        object_positions : `np.ndarray`
            Pre-computed RA,dec positions for each object in orbits (degrees)
        object_mjds : `np.ndarray`
            MJD values for each pre-computed position
        """

        # If we are trying to use pre-computed positions, check that the MJDs span enough
        if object_mjds is not None:
            if (obs_data[self.obs_time_col].min() < object_mjds.min()) | (
                obs_data[self.obs_time_col].max() > object_mjds.max()
            ):
                warnings.warn(
                    "Pre-computed position times do not cover MJD range of %i-%i."
                    % (obs_data[self.obs_time_col].min(), obs_data[self.obs_time_col].max())
                )
                object_mjds = None
            # calculate angular motion for each object at each timestep
            # how much does it move going to time step forward
            move1 = angular_separation(
                object_positions["ra"][:, 0:-1],
                object_positions["dec"][:, 0:-1],
                object_positions["ra"][:, 1:],
                object_positions["dec"][:, 1:],
            )
            # Need to take care of ends
            d1 = object_positions["ra"] * 0
            d1[:, 0:-1] = move1
            d2 = object_positions["ra"] * 0
            d2[:, 1:] = move1
            # Now we can use a small tolerance for objects when they are moving
            # slowly across the sky.
            object_tol = np.maximum(d1, d2) + self.pre_comp_tol

        # output dtype
        names = [
            "obj_id",
            self.obs_id_col,
            "sedname",
            "time",
            "ra",
            "dec",
            "dradt",
            "ddecdt",
            "phase",
            "solarelon",
            "helio_dist",
            "geo_dist",
            "magV",
            "trueAnomaly",
            "velocity",
            "dmag_color",
            "dmag_trail",
            "dmag_detect",
        ]
        types = [int, int, "<U40"] + [float] * (len(names) - 3)

        # XXX--for now, copy over some observation info. In the future,
        # just index it to the observations and only transfer observationId
        transfer_cols = [
            self.visit_exp_time_col,
            self.m5_col,
            self.seeing_col,
            self.obs_time_col,
            self.night_col,
            self.filter_col,
        ]
        transfer_types = [float] * (len(transfer_cols) - 2)
        transfer_types += [int, "<U40"]
        names += transfer_cols
        types += transfer_types

        output_dtype = list(zip(names, types))

        # Set the times for the rough ephemeris grid.
        time_step = float(self.tstep)
        time_start = np.floor(obs_data[self.obs_time_col].min() + 0.16 - 0.5) - time_step
        time_end = np.ceil(obs_data[self.obs_time_col].max() + 0.16 + 0.5) + time_step
        rough_times = np.arange(time_start, time_end + time_step / 2.0, time_step)
        if self.verbose:
            logging.info("Generating preliminary ephemerides on a grid of %f day timesteps." % (time_step))
        # list to hold results for each object
        result = []
        # save indx to match observation indx to object indx
        indx_map_visit_to_object = []
        # For each object, identify observations where the object is within the FOV (or camera footprint).
        for i, sso in enumerate(orbits):
            objid = sso.orbits["obj_id"].iloc[0]
            sedname = sso.orbits["sed_filename"].iloc[0]
            # Generate ephemerides on the rough grid.
            if self.verbose:
                logging.debug(
                    ("%d/%d   id=%s : " % (i, len(orbits), objid))
                    + datetime.datetime.now().strftime("Prelim start: %Y-%m-%d %H:%M:%S")
                    + " nRoughTimes: %s" % len(rough_times)
                )
            # Not using pre-computed positions
            if object_mjds is None:
                ephs = self.generate_ephemerides(
                    sso,
                    rough_times,
                    eph_mode=self.prelim_eph_mode,
                    eph_type=self.eph_type,
                )[0]
                mu = ephs["velocity"]
                if self.verbose:
                    logging.debug(
                        ("%d/%d   id=%s : " % (i, len(orbits), objid))
                        + datetime.datetime.now().strftime("Prelim end: %Y-%m-%d %H:%M:%S")
                        + " π(median, max), min(geo_dist): %.2f, %.2f deg/day  %.2f AU"
                        % (np.median(mu), np.max(mu), np.min(ephs["geo_dist"]))
                    )

                # Find observations which come within roughTol of the fov.
                ephs_idxs = np.searchsorted(ephs["time"], obs_data[self.obs_time_col])
                rough_idx_obs = self._sso_in_circle_fov(ephs[ephs_idxs], obs_data, self.rough_tol)
            else:
                # Nearest neighbor search for the object_mjd closest to obs_data mjd
                pos = np.searchsorted(object_mjds, obs_data[self.obs_time_col], side="left")
                pos_right = pos - 1
                object_indx = pos + 0
                d_left = obs_data[self.obs_time_col] - object_mjds[pos]
                d_right = obs_data[self.obs_time_col] - object_mjds[pos_right]
                r_better = np.where(np.abs(d_right) < np.abs(d_left))[0]
                object_indx[r_better] = pos_right[r_better]

                rough_idx_obs = self._sso_in_circle_fov(
                    object_positions[i][object_indx],
                    obs_data,
                    self.r_fov + object_tol[i][object_indx],
                )
            if len(rough_idx_obs) > 0:
                # Generate exact ephemerides for these times.
                times = obs_data[self.obs_time_col][rough_idx_obs]
                if self.verbose:
                    logging.debug(
                        ("%d/%d   id=%s : " % (i, len(orbits), objid))
                        + datetime.datetime.now().strftime("Exact start: %Y-%m-%d %H:%M:%S")
                        + " nExactTimes: %s" % len(times)
                    )
                ephs = self.generate_ephemerides(sso, times, eph_mode=self.eph_mode, eph_type=self.eph_type)[
                    0
                ]
                logging.debug(
                    ("%d/%d   id=%s : " % (i, len(orbits), objid))
                    + datetime.datetime.now().strftime("Exact end: %Y-%m-%d %H:%M:%S")
                )
                # Identify the objects which fell within the specific footprint.
                idx_obs = self.sso_in_fov(ephs, obs_data[rough_idx_obs])
                if self.verbose:
                    logging.info(
                        ("%d/%d   id=%s : " % (i, len(orbits), objid))
                        + "Object in %d out of %d potential fields (%.2f%% success rate)"
                        % (
                            len(idx_obs),
                            len(times),
                            100.0 * float(len(idx_obs)) / len(times),
                        )
                    )
                object_observations = np.zeros(idx_obs.size, dtype=output_dtype)
                object_observations["obj_id"] = objid
                object_observations[self.obs_id_col] = obs_data[rough_idx_obs][idx_obs][self.obs_id_col]
                object_observations["sedname"] = sedname

                for key in ephs.dtype.names:
                    object_observations[key] = ephs[key][idx_obs].copy()
                result.append(object_observations)
                indx_map_visit_to_object.append(rough_idx_obs[idx_obs])

        if len(result) > 0:
            result = np.concatenate(result)
            indx_map_visit_to_object = np.concatenate(indx_map_visit_to_object)
            # add on any additional info we want here, dmags, etc
            filterlist = np.unique(obs_data[self.filter_col][indx_map_visit_to_object])
            for sname in np.unique(result["sedname"]):
                dmag_color_dict = self.calc_colors(sname)
                for f in filterlist:
                    match = np.where(
                        (obs_data[self.filter_col][indx_map_visit_to_object] == f)
                        & (result["sedname"] == sname)
                    )[0]
                    if np.size(match) > 0:
                        result["dmag_color"][match] = dmag_color_dict[f]
            # Calculate trailing and detection loses.
            result["dmag_trail"], result["dmag_detect"] = self.calc_trailing_losses(
                result["velocity"],
                obs_data[self.seeing_col][indx_map_visit_to_object],
                obs_data[self.visit_exp_time_col][indx_map_visit_to_object],
            )

            # Transfer over info from pointing info array to result array
            for key in transfer_cols:
                result[key] = obs_data[key][indx_map_visit_to_object]

        return result



import unittest

import numpy as np

import rubin_sim.maf.metrics as metrics


class TestStringCount(unittest.TestCase):
    def testsc(self):
        metric = metrics.StringCountMetric()
        data = np.array(["a", "a", "b", "c", "", "", ""])
        dt = np.dtype([("filter", np.str_, 1)])
        data.dtype = dt
        result = metric.run(data)
        # Check that the metricValue is correct
        expected_results = {"a": 2, "b": 1, "c": 1, "blank": 3}
        for key in expected_results:
            assert result[key] == expected_results[key]

        # Check that the reduce functions got made and return expected result
        for key in expected_results:
            assert metric.reduce_funcs[key](result) == expected_results[key]


if __name__ == "__main__":
    unittest.main()



import unittest
from tempfile import TemporaryFile

import healpy as hp
import numpy as np
import numpy.ma as ma

import rubin_sim.maf.slicers as slicers


class TestSlicers(unittest.TestCase):
    def setUp(self):
        self.filenames = []
        self.baseslicer = slicers.BaseSlicer()

    def test_healpix_slicer_obj(self):
        rng = np.random.RandomState(8121)
        nside = 32
        slicer = slicers.HealpixSlicer(nside=nside)
        metric_values = rng.rand(hp.nside2npix(nside)).astype("object")
        metric_values = ma.MaskedArray(
            data=metric_values,
            mask=np.where(metric_values < 0.1, True, False),
            fill_value=slicer.badval,
        )
        with TemporaryFile() as filename:
            info_label = "testdata"
            slicer.write_data(filename, metric_values, info_label=info_label)
            _ = filename.seek(0)
            metric_values_back, slicer_back, header = self.baseslicer.read_data(filename)
            np.testing.assert_almost_equal(metric_values_back, metric_values)
            assert slicer == slicer_back
            assert info_label == header["info_label"]
            attr2check = ["nside", "nslice", "columns_needed", "lon_col", "lat_col"]
            for att in attr2check:
                assert getattr(slicer, att) == getattr(slicer_back, att)

    def test_healpix_slicer_floats(self):
        rng = np.random.RandomState(71231)
        nside = 32
        slicer = slicers.HealpixSlicer(nside=nside)
        metric_values = rng.rand(hp.nside2npix(nside))
        with TemporaryFile() as filename:
            slicer.write_data(filename, metric_values, info_label="testdata")
            _ = filename.seek(0)
            metric_values_back, slicer_back, header = self.baseslicer.read_data(filename)
            np.testing.assert_almost_equal(metric_values_back, metric_values)
            assert slicer == slicer_back
            attr2check = ["nside", "nslice", "columns_needed", "lon_col", "lat_col"]
            for att in attr2check:
                assert getattr(slicer, att) == getattr(slicer_back, att)

    def test_healpix_slicer_masked(self):
        rng = np.random.RandomState(712551)
        nside = 32
        slicer = slicers.HealpixSlicer(nside=nside)
        metric_values = rng.rand(hp.nside2npix(nside))
        metric_values = ma.MaskedArray(
            data=metric_values,
            mask=np.where(metric_values < 0.1, True, False),
            fill_value=slicer.badval,
        )
        with TemporaryFile() as filename:
            slicer.write_data(filename, metric_values, info_label="testdata")
            _ = filename.seek(0)
            metric_values_back, slicer_back, header = self.baseslicer.read_data(filename)
            np.testing.assert_almost_equal(metric_values_back, metric_values)
            assert slicer == slicer_back
            attr2check = ["nside", "nslice", "columns_needed", "lon_col", "lat_col"]
            for att in attr2check:
                assert getattr(slicer, att) == getattr(slicer_back, att)

    def test_one_d_slicer(self):
        rng = np.random.RandomState(71111)
        slicer = slicers.OneDSlicer(slice_col_name="testdata")
        data_values = np.zeros(10000, dtype=[("testdata", "float")])
        data_values["testdata"] = rng.rand(10000)
        slicer.setup_slicer(data_values)
        with TemporaryFile() as filename:
            slicer.write_data(filename, data_values[:100])
            _ = filename.seek(0)
            data_back, slicer_back, header = self.baseslicer.read_data(filename)
            assert slicer == slicer_back
            # np.testing.assert_almost_equal(data_back,data_values[:100])
            attr2check = ["nslice", "columns_needed"]
            for att in attr2check:
                if type(getattr(slicer, att)).__module__ == "numpy":
                    np.testing.assert_almost_equal(getattr(slicer, att), getattr(slicer_back, att))
                else:
                    assert getattr(slicer, att) == getattr(slicer_back, att)

    def test_userpoints_slicer(self):
        rng = np.random.RandomState(7442)

        names = ["fieldRA", "fieldDec", "fieldId"]
        dt = ["float", "float", "int"]
        metric_values = rng.rand(100)
        field_data = np.zeros(100, dtype=list(zip(names, dt)))
        field_data["fieldRA"] = rng.rand(100)
        field_data["fieldDec"] = rng.rand(100)
        field_data["fieldId"] = np.arange(100)

        slicer = slicers.UserPointsSlicer(ra=field_data["fieldRA"], dec=field_data["fieldDec"])

        names = ["data1", "data2", "fieldId"]
        sim_data = np.zeros(100, dtype=list(zip(names, dt)))
        sim_data["data1"] = rng.rand(100)
        sim_data["fieldId"] = np.arange(100)
        with TemporaryFile() as filename:
            slicer.write_data(filename, metric_values)
            _ = filename.seek(0)
            metric_back, slicer_back, header = self.baseslicer.read_data(filename)
            assert slicer == slicer_back
            np.testing.assert_almost_equal(metric_back, metric_values)
            attr2check = [
                "nslice",
                "columns_needed",
                "lon_col",
                "lat_col",
                "shape",
            ]
            for att in attr2check:
                if type(getattr(slicer, att)).__name__ == "dict":
                    for key in getattr(slicer, att):
                        np.testing.assert_almost_equal(
                            getattr(slicer, att)[key], getattr(slicer_back, att)[key]
                        )
                else:
                    assert getattr(slicer, att) == getattr(slicer_back, att)

    def test_unislicer(self):
        rng = np.random.RandomState(34229)
        slicer = slicers.UniSlicer()
        data = np.zeros(1, dtype=[("testdata", "float")])
        data[:] = rng.rand(1)
        slicer.setup_slicer(data)
        with TemporaryFile() as filename:
            metric_value = np.array([25.0])
            slicer.write_data(filename, metric_value)
            _ = filename.seek(0)
            data_back, slicer_back, header = self.baseslicer.read_data(filename)
            assert slicer == slicer_back
            np.testing.assert_almost_equal(data_back, metric_value)
            attr2check = ["nslice", "columns_needed"]
            for att in attr2check:
                assert getattr(slicer, att) == getattr(slicer_back, att)

    def test_complex(self):
        # Test case where there is a complex metric
        rng = np.random.RandomState(5442)
        nside = 8
        slicer = slicers.HealpixSlicer(nside=nside)
        data = np.zeros(slicer.nslice, dtype="object")
        for i, ack in enumerate(data):
            n_el = rng.rand(1) * 4  # up to 4 elements
            data[i] = np.arange(n_el)
        with TemporaryFile() as filename:
            slicer.write_data(filename, data)
            _ = filename.seek(0)
            data_back, slicer_back, header = self.baseslicer.read_data(filename)
            assert slicer == slicer_back
            # This is a crazy slow loop!
            for i, ack in enumerate(data):
                np.testing.assert_almost_equal(data_back[i], data[i])

    def test_n_d_slicer(self):
        rng = np.random.RandomState(621)
        colnames = ["test1", "test2", "test3"]
        data = []
        for c in colnames:
            data.append(rng.rand(1000))
        dv = np.core.records.fromarrays(data, names=colnames)
        slicer = slicers.NDSlicer(colnames, bins_list=10)
        slicer.setup_slicer(dv)
        with TemporaryFile() as filename:
            metricdata = np.zeros(slicer.nslice, dtype="float")
            for i, s in enumerate(slicer):
                metricdata[i] = i
            slicer.write_data(filename, metricdata)
            _ = filename.seek(0)
            data_back, slicer_back, header = self.baseslicer.read_data(filename)
            assert slicer == slicer_back
            np.testing.assert_almost_equal(data_back, metricdata)


if __name__ == "__main__":
    unittest.main()



__all__ = ("GRBTransientMetric",)


import numpy as np

import rubin_sim.maf.metrics as metrics

# Gamma-ray burst afterglow metric
# ebellm@caltech.edu


class GRBTransientMetric(metrics.BaseMetric):
    """Detections for on-axis GRB afterglows decaying as
        F(t) = F(1min)((t-t0)/1min)^-alpha.  No jet break, for now.

        Derived from TransientMetric, but calculated with reduce functions to
    enable-band specific counts.
        Burst parameters taken from 2011PASP..123.1034J.

        Simplifications:
        no color variation or evolution encoded yet.
        no jet breaks.
        not treating off-axis events.

    Parameters
    ----------
    alpha : float,
        temporal decay index
        Default = 1.0
    apparent_mag_1min_mean : float,
        mean magnitude at 1 minute after burst
        Default = 15.35
    apparent_mag_1min_sigma : float,
        std of magnitudes at 1 minute after burst
        Default = 1.59
    trans_duration : float, optional
        How long the transient lasts (days). Default 10.
    survey_duration : float, optional
        Length of survey (years).
        Default 10.
    survey_start : float, optional
        MJD for the survey start date.
        Default None (uses the time of the first observation).
    detect_m5_plus : float, optional
        An observation will be used if the light curve magnitude is brighter than m5+detect_m5_plus.
        Default 0.
    n_per_filter : int, optional
        Number of separate detections of the light curve above the
        detect_m5_plus theshold (in a single filter) for the light curve
        to be counted.
        Default 1.
    n_filters : int, optional
        Number of filters that need to be observed n_per_filter times,
        with differences min_delta_mag,
        for an object to be counted as detected.
        Default 1.
    min_delta_mag : float, optional
       magnitude difference between detections in the same filter required
       for second+ detection to be counted.
       For example, if min_delta_mag = 0.1 mag and two consecutive observations
       differ only by 0.05 mag, those two detections will only count as one.
       (Better would be a SNR-based discrimination of lightcurve change.)
       Default 0.
    n_phase_check : int, optional
        Sets the number of phases that should be checked.
        One can imagine pathological cadences where many objects pass the detection criteria,
        but would not if the observations were offset by a phase-shift.
        Default 1.
    """

    def __init__(
        self,
        alpha=1,
        apparent_mag_1min_mean=15.35,
        apparent_mag_1min_sigma=1.59,
        metric_name="GRBTransientMetric",
        mjd_col="expMJD",
        m5_col="fiveSigmaDepth",
        filter_col="filter",
        trans_duration=10.0,
        survey_duration=10.0,
        survey_start=None,
        detect_m5_plus=0.0,
        n_per_filter=1,
        n_filters=1,
        min_delta_mag=0.0,
        n_phase_check=1,
        **kwargs,
    ):
        self.mjd_col = mjd_col
        self.m5_col = m5_col
        self.filter_col = filter_col
        super(GRBTransientMetric, self).__init__(
            col=[self.mjd_col, self.m5_col, self.filter_col],
            units="Fraction Detected",
            metric_name=metric_name,
            **kwargs,
        )
        self.alpha = alpha
        self.apparent_mag_1min_mean = apparent_mag_1min_mean
        self.apparent_mag_1min_sigma = apparent_mag_1min_sigma
        self.trans_duration = trans_duration
        self.survey_duration = survey_duration
        self.survey_start = survey_start
        self.detect_m5_plus = detect_m5_plus
        self.n_per_filter = n_per_filter
        self.n_filters = n_filters
        self.min_delta_mag = min_delta_mag
        self.n_phase_check = n_phase_check
        self.peak_time = 0.0
        self.reduce_order = {
            "Bandu": 0,
            "Bandg": 1,
            "Bandr": 2,
            "Bandi": 3,
            "Bandz": 4,
            "Bandy": 5,
            "Band1FiltAvg": 6,
            "BandanyNfilters": 7,
        }

    def light_curve(self, time, filters):
        """
        given the times and filters of an observation, return the magnitudes.
        """

        lc_mags = np.zeros(time.size, dtype=float)

        decline = np.where(time > self.peak_time)
        apparent_mag_1min = np.random.randn() * self.apparent_mag_1min_sigma + self.apparent_mag_1min_mean
        lc_mags[decline] += apparent_mag_1min + self.alpha * 2.5 * np.log10(
            (time[decline] - self.peak_time) * 24.0 * 60.0
        )

        # for key in self.peaks.keys():
        #    fMatch = np.where(filters == key)
        #    lc_mags[fMatch] += self.peaks[key]

        return lc_mags

    def run(self, data_slice, slice_point=None):
        """ "
        Calculate the detectability of a transient with the specified lightcurve.

        Parameters
        ----------
        data_slice : numpy.array
            Numpy structured array containing the data related to the visits provided by the slicer.
        slice_point : dict, optional
            Dictionary containing information about the slice_point currently active in the slicer.

        Returns
        -------
        float
            The total number of transients that could be detected.
        """
        # Total number of transients that could go off back-to-back
        n_trans_max = np.floor(self.survey_duration / (self.trans_duration / 365.25))
        tshifts = np.arange(self.n_phase_check) * self.trans_duration / float(self.n_phase_check)
        n_trans_max = 0
        for tshift in tshifts:
            # Compute the total number of back-to-back transients are possible to detect
            # given the survey duration and the transient duration.
            n_trans_max += np.floor(self.survey_duration / (self.trans_duration / 365.25))
            if tshift != 0:
                n_trans_max -= 1
            if self.survey_start is None:
                survey_start = data_slice[self.mjd_col].min()
            time = (data_slice[self.mjd_col] - survey_start + tshift) % self.trans_duration

            # Which lightcurve does each point belong to
            lc_number = np.floor((data_slice[self.mjd_col] - survey_start) / self.trans_duration)

            lc_mags = self.light_curve(time, data_slice[self.filter_col])

            # How many criteria needs to be passed
            detect_thresh = 0

            # Flag points that are above the SNR limit
            detected = np.zeros(data_slice.size, dtype=int)
            detected[np.where(lc_mags < data_slice[self.m5_col] + self.detect_m5_plus)] += 1

            bandcounter = {
                "u": 0,
                "g": 0,
                "r": 0,
                "i": 0,
                "z": 0,
                "y": 0,
                "any": 0,
            }  # define zeroed out counter

            # make sure things are sorted by time
            ord = np.argsort(data_slice[self.mjd_col])
            data_slice = data_slice[ord]
            detected = detected[ord]
            lc_number = lc_number[ord]
            lc_mags = lc_mags[ord]
            ulc_number = np.unique(lc_number)
            left = np.searchsorted(lc_number, ulc_number)
            right = np.searchsorted(lc_number, ulc_number, side="right")
            detect_thresh += self.n_filters

            # iterate over the lightcurves
            for le, ri in zip(left, right):
                wdet = np.where(detected[le:ri] > 0)
                ufilters = np.unique(data_slice[self.filter_col][le:ri][wdet])
                nfilts_lci = 0
                for filt_name in ufilters:
                    wdetfilt = np.where((data_slice[self.filter_col][le:ri] == filt_name) & detected[le:ri])

                    lc_points = lc_mags[le:ri][wdetfilt]
                    dlc = np.abs(np.diff(lc_points))

                    # number of detections in band, requring that for
                    # nPerFilter > 1 that points have more than minDeltaMag
                    # change
                    nbanddet = np.sum(dlc > self.min_delta_mag) + 1
                    if nbanddet >= self.n_per_filter:
                        bandcounter[filt_name] += 1
                        nfilts_lci += 1
                if nfilts_lci >= self.n_filters:
                    bandcounter["any"] += 1

        bandfraction = {}
        for band in bandcounter.keys():
            bandfraction[band] = float(bandcounter[band]) / n_trans_max

        return bandfraction

    def reduce_band1_filt_avg(self, bandfraction):
        "Average fraction detected in single filter"
        return np.mean(list(bandfraction.values()))

    def reduce_bandany_nfilters(self, bandfraction):
        "Fraction of events detected in Nfilters or more"
        return bandfraction["any"]

    def reduce_bandu(self, bandfraction):
        return bandfraction["u"]

    def reduce_bandg(self, bandfraction):
        return bandfraction["g"]

    def reduce_bandr(self, bandfraction):
        return bandfraction["r"]

    def reduce_bandi(self, bandfraction):
        return bandfraction["i"]

    def reduce_bandz(self, bandfraction):
        return bandfraction["z"]

    def reduce_bandy(self, bandfraction):
        return bandfraction["y"]



__all__ = (
    "BaseMoStacker",
    "MoMagStacker",
    "AppMagStacker",
    "CometAppMagStacker",
    "SNRStacker",
    "EclStacker",
)

import warnings

import numpy as np

from .base_stacker import BaseStacker
from .mo_phase import phase__halley_marcus

# Willmer 2018, ApJS 236, 47
# VEGA V mag and AB mag of sun (LSST-equivalent bandpasses)
VMAG_SUN = -26.76  # Vega mag
AB_SUN = {"u": -25.30, "g": -26.52, "r": -26.93, "i": -27.05, "z": -27.07, "y": -27.07}
KM_PER_AU = 149597870.7


class BaseMoStacker(BaseStacker):
    """Base class for moving object (SSobject)  stackers. Relevant for MoSlicer ssObs (pd.dataframe).

    Provided to add moving-object specific API for 'run' method of moving object stackers.
    """

    def run(self, sso_obs, href, hval=None):
        # Redefine this here, as the API does not match BaseStacker.
        if hval is None:
            hval = href
        if len(sso_obs) == 0:
            return sso_obs
        # Add the columns.
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            sso_obs, cols_present = self._add_stacker_cols(sso_obs)
        # Here we don't really care about cols_present, because almost every time we will be readding
        # columns anymore (for different H values).
        return self._run(sso_obs, href, hval)


class MoMagStacker(BaseMoStacker):
    """Add columns relevant to SSobject apparent magnitudes and visibility to the slicer ssoObs
    dataframe, given a particular Href and current h_val.

    Specifically, this stacker adds magLimit, appMag, SNR, and vis.
    magLimit indicates the appropriate limiting magnitude to consider for a particular object in a particular
    observation, when combined with the losses due to detection (dmag_detect) or trailing (dmagTrail).
    appMag adds the apparent magnitude in the filter of the current object, at the current h_val.
    SNR adds the SNR of this object, given the magLimit.
    vis adds a flag (0/1) indicating whether an object was visible (assuming a 5sigma threshhold including
    some probabilistic determination of visibility).

    Parameters
    ----------
    m5Col : `str`, optional
        Name of the column describing the 5 sigma depth of each visit. Default fiveSigmaDepth.
    lossCol : `str`, optional
        Name of the column describing the magnitude losses,
        due to trailing (dmagTrail) or detection (dmag_detect). Default dmag_detect.
    gamma : `float`, optional
        The 'gamma' value for calculating SNR. Default 0.038.
        LSST range under normal conditions is about 0.037 to 0.039.
    sigma : `float`, optional
        The 'sigma' value for probabilistic prediction of whether or not an object is visible at 5sigma.
        Default 0.12.
        The probabilistic prediction of visibility is based on Fermi-Dirac completeness formula (see SDSS,
        eqn 24, Stripe82 analysis: http://iopscience.iop.org/0004-637X/794/2/120/pdf/apj_794_2_120.pdf).
    randomSeed: `int` or None, optional
        If set, then used as the random seed for the numpy random number
        generation for the dither offsets.
        Default: None.
    """

    cols_added = ["appMag", "SNR", "vis"]

    def __init__(
        self,
        magtype="asteroid",
        v_mag_col="magV",
        color_col="dmag_color",
        loss_col="dmag_detect",
        m5_col="fiveSigmaDepth",
        seeing_col="seeingFwhmGeom",
        filter_col="filter",
        gamma=0.038,
        sigma=0.12,
        random_seed=None,
    ):
        if magtype == "asteroid":
            self.mag_stacker = AppMagStacker(v_mag_col=v_mag_col, color_col=color_col, loss_col=loss_col)
            self.cols_req = [m5_col, v_mag_col, color_col, loss_col]
        elif magtype.startswith("comet"):
            # magtype should be = comet_oort comet_short or comet_mbc
            comet_type = magtype.split(magtype, "_")[-1]
            self.mag_stacker = CometAppMagStacker(
                comet_type=comet_type,
                ap=0.04,
                rh_col="helio_dist",
                delta_col="geo_dist",
                phase_col="phase",
                seeing_col=seeing_col,
                ap_scale=1,
                filter_col=filter_col,
                v_mag_col=v_mag_col,
                color_col=color_col,
                loss_col=loss_col,
            )
            self.cols_req = [
                m5_col,
                v_mag_col,
                color_col,
                loss_col,
                "helio_dist",
                "geo_dist",
                "phase",
                seeing_col,
                filter_col,
            ]
        else:
            self.mag_stacker = AppMagNullStacker()
        self.snr_stacker = SNRStacker(
            app_mag_col="appMag",
            m5_col=m5_col,
            gamma=gamma,
            sigma=sigma,
            random_seed=random_seed,
        )

        self.units = ["mag", "mag", "SNR", ""]

    def _run(self, sso_obs, href, hval):
        # hval = current H value (useful if cloning over H range), href = reference H value from orbit.
        # Without cloning, href = hval.
        # add apparent magnitude
        self.mag_stacker._run(sso_obs, href, hval)
        # add snr
        self.snr_stacker._run(sso_obs, href, hval)
        return sso_obs


class AppMagNullStacker(BaseMoStacker):
    """Do nothing to calculate an apparent magnitude.

    This assumes an apparent magnitude was part of the input data and does not need to be modified (no
    cloning, color terms, trailing losses, etc). Just return the appMag column.

    This would not be necessary in general, but appMag is treated as a special column (because we must have an
    apparent magnitude for most of the basic moving object metrics, and it must be calculated before SNR
    if that is also needed).
    """

    cols_added = ["appMag"]

    def __init__(self, app_mag_col="appMag"):
        self.app_mag_col = app_mag_col
        self.units = [
            "mag",
        ]
        self.cols_req = [self.app_mag_col]

    def _run(self, sso_obs, href, hval):
        sso_obs["appMag"] = sso_obs[self.app_mag_col]
        return sso_obs


class AppMagStacker(BaseMoStacker):
    """Add apparent magnitude of an object for the current h_val (compared to Href in the orbit file),
    incorporating the magnitude losses due to trailing/detection, as well as the color of the object.

    This is calculated from the reported mag_v in the input observation file (calculated assuming Href) as:
    ssoObs['appMag'] = ssoObs[self.vMagCol] + ssoObs[self.colorCol] + ssoObs[self.lossCol] + h_val - Href

    Using the vMag reported in the input observations implicitly uses the phase curve coded in at that point;
    for Oorb this is an H/G phase curve, with G=0.15 unless otherwise specified in the orbit file.
    See sims_movingObjects for more details on the color and loss quantities.

    Parameters
    ----------
    v_mag_col : `str`, optional
        Name of the column containing the base V magnitude for the object at H=Href.
    loss_col : `str`, optional
        Name of the column describing the magnitude losses,
        due to trailing (dmagTrail) or detection (dmag_detect). Default dmag_detect.
    color_col : `str`, optional
        Name of the column describing the color correction (into the observation filter, from V).
        Default dmag_color.
    """

    cols_added = ["appMag"]

    def __init__(self, v_mag_col="magV", color_col="dmag_color", loss_col="dmag_detect"):
        self.v_mag_col = v_mag_col
        self.color_col = color_col
        self.loss_col = loss_col
        self.cols_req = [self.v_mag_col, self.color_col, self.loss_col]
        self.units = [
            "mag",
        ]

    def _run(self, sso_obs, href, hval):
        # hval = current H value (useful if cloning over H range), href = reference H value from orbit.
        # Without cloning, href = hval.
        sso_obs["appMag"] = (
            sso_obs[self.v_mag_col] + sso_obs[self.color_col] + sso_obs[self.loss_col] + hval - href
        )
        return sso_obs


class CometAppMagStacker(BaseMoStacker):
    """Add a cometary apparent magnitude, including nucleus and coma, based on a calculation of
    Afrho (using the current h_val) and a Halley-Marcus phase curve for the coma brightness.

    Parameters
    ----------
    cometType : `str`, optional
        Type of comet - short, oort, or mbc. This setting also sets the value of Afrho1 and k:
        short = Afrho1 / R^2 = 100 cm/km2, k = -4
        oort = Afrho1 / R^2 = 1000 cm/km2, k = -2
        mbc = Afrho1 / R^2 = 4000 cm/km2, k = -6.
        Default = 'oort'.
        It is also possible to pass this a dictionary instead: the dictionary should contain 'k' and
        'afrho1_const' keys, which will be used to set these values directly.
        (e.g. cometType = {'k': -3.5, 'afrho1_const': 1500}).
    ap : `float`, optional
        The albedo for calculating the object's size. Default 0.04
    rh_col : `str`, optional
        The column name for the heliocentric distance (in AU). Default 'helio_dist'.
    delta_col : `str`, optional
        The column name for the geocentric distance (in AU). Default 'geo_dist'.
    phase_col : `str`, optional
        The column name for the phase value (in degrees). Default 'phase'.
    """

    cols_added = ["appMag"]

    def __init__(
        self,
        comet_type="oort",
        ap=0.04,
        rh_col="helio_dist",
        delta_col="geo_dist",
        phase_col="phase",
        seeing_col="FWHMgeom",
        ap_scale=1,
        filter_col="filter",
        v_mag_col="magV",
        color_col="dmag_color",
        loss_col="dmag_detect",
    ):
        self.units = ["mag"]  # new column units
        # Set up k and Afrho1 constant values.
        comet_types = {
            "short": {"k": -4, "Afrho1_const": 100},
            "oort": {"k": -2, "Afrho1_const": 1000},
            "mbc": {"k": -6, "Afrho1_const": 4000},
        }
        self.k = None
        self.afrho1_const = None
        if isinstance(comet_type, str):
            if comet_type in comet_types:
                self.k = comet_types[comet_type]["k"]
                self.afrho1_const = comet_types[comet_type]["Afrho1_const"]
        if isinstance(comet_type, dict):
            if "k" in comet_type:
                self.k = comet_type["k"]
            if "Afrho1_const" in comet_type:
                self.afrho1_const = comet_type["Afrho1_const"]
        if self.k is None or self.afrho1_const is None:
            raise ValueError(
                f"comet_type must be a string {comet_types} or "
                f'dict containing "k" and "Afrho1_const" - but received {comet_type}'
            )
        # Phew, now set the simple stuff.
        self.ap = ap
        self.rh_col = rh_col
        self.delta_col = delta_col
        self.phase_col = phase_col
        self.seeing_col = seeing_col
        self.ap_scale = ap_scale
        self.filter_col = filter_col
        self.v_mag_col = v_mag_col
        self.color_col = color_col
        self.loss_col = loss_col
        # names of required columns
        self.cols_req = [
            self.rh_col,
            self.delta_col,
            self.phase_col,
            self.seeing_col,
            self.filter_col,
            self.v_mag_col,
            self.color_col,
            self.loss_col,
        ]

    def _run(self, sso_obs, href, hval):
        # Calculate radius from the current H value (hval).
        radius = 10 ** (0.2 * (VMAG_SUN - hval)) / np.sqrt(self.ap) * KM_PER_AU
        # Calculate expected Afrho - this is a value that describes how the brightness of the coma changes
        afrho1 = self.afrho1_const * radius**2
        phase_val = phase__halley_marcus(sso_obs[self.phase_col])
        # afrho is equivalent to a sort of 'absolute' magnitude of the coma
        afrho = afrho1 * sso_obs[self.rh_col] ** self.k * phase_val
        # rho describes the projected area on the sky (project the aperture into cm on-sky)
        # Using the seeing * apScale as the aperture
        radius_aperture = sso_obs[self.seeing_col] * self.ap_scale
        rho = 725e5 * sso_obs[self.delta_col] * radius_aperture
        # Calculate the expected apparent of the comet coma = sun + correction
        delta = sso_obs[self.delta_col] * KM_PER_AU * 1000  # delta in cm
        dm = -2.5 * np.log10(afrho * rho / (2 * sso_obs[self.rh_col] * delta) ** 2)
        coma = np.zeros(len(sso_obs), float)
        # This calculates a coma mag that scales with the sun's color in each bandpass, but the coma
        # modification is gray (i.e. it's just reflecting sunlight)
        for f in sso_obs[self.filter_col]:
            match = np.where(sso_obs[self.filter_col] == f)
            coma[match] = AB_SUN[f] + dm[match]
        # Calculate cometary nucleus magnitude -- we'll use the apparent V mag adapted from OOrb as well as
        # the object's color - these are generally assumed to be D type (which was used in sims_movingObjects)
        nucleus = sso_obs[self.v_mag_col] + sso_obs[self.color_col] + sso_obs[self.loss_col] + hval - href
        # add coma and nucleus then ready for calculation of SNR, etc.
        sso_obs["appMag"] = -2.5 * np.log10(10 ** (-0.4 * coma) + 10 ** (-0.4 * nucleus))
        return sso_obs


class SNRStacker(BaseMoStacker):
    """Add SNR and visibility for a particular object, given the five sigma depth of the image and the
    apparent magnitude (whether from AppMagStacker or CometAppMagStacker, etc).

    The SNR simply calculates the SNR based on the five sigma depth and the apparent magnitude.
    The 'vis' column is a probabilistic flag (0/1) indicating whether the object was detected, assuming
    a 5-sigma SNR threshold and then applying a probabilistic cut on whether it was detected or not (i.e.
    there is a gentle roll-over in 'vis' from 1 to 0 depending on the SNR of the object).
    This is based on the Fermi-Dirac completeness formula as described in equation 24 of the Stripe 82 SDSS
    analysis here: http://iopscience.iop.org/0004-637X/794/2/120/pdf/apj_794_2_120.pdf.

    Parameters
    ----------
    app_mag_col : `str`, optional
        Name of the column describing the apparent magnitude of the object. Default 'appMag'.
    m5_col : `str`, optional
        Name of the column describing the 5 sigma depth of each visit. Default fiveSigmaDepth.
    gamma : `float`, optional
        The 'gamma' value for calculating SNR. Default 0.038.
        LSST range under normal conditions is about 0.037 to 0.039.
    sigma : `float`, optional
        The 'sigma' value for probabilistic prediction of whether or not an object is visible at 5sigma.
        Default 0.12.
        The probabilistic prediction of visibility is based on Fermi-Dirac completeness formula (see SDSS,
        eqn 24, Stripe82 analysis: http://iopscience.iop.org/0004-637X/794/2/120/pdf/apj_794_2_120.pdf).
    random_seed: `int` or None, optional
        If set, then used as the random seed for the numpy random number
        generation for the probability of detection. Default: None.
    """

    cols_added = ["SNR", "vis"]

    def __init__(
        self,
        app_mag_col="appMag",
        m5_col="fiveSigmaDepth",
        gamma=0.038,
        sigma=0.12,
        random_seed=None,
    ):
        self.app_mag_col = app_mag_col
        self.m5_col = m5_col
        self.gamma = gamma
        self.sigma = sigma
        self.random_seed = random_seed
        self.cols_req = [self.app_mag_col, self.m5_col]
        self.units = ["SNR", ""]

    def _run(self, sso_obs, href, hval):
        # hval = current H value (useful if cloning over H range), href = reference H value from orbit.
        # Without cloning, href = hval.
        xval = np.power(10, 0.5 * (sso_obs[self.app_mag_col] - sso_obs[self.m5_col]))
        sso_obs["SNR"] = 1.0 / np.sqrt((0.04 - self.gamma) * xval + self.gamma * xval * xval)
        completeness = 1.0 / (1 + np.exp((sso_obs[self.app_mag_col] - sso_obs[self.m5_col]) / self.sigma))
        if not hasattr(self, "_rng"):
            if self.random_seed is not None:
                self._rng = np.random.RandomState(self.random_seed)
            else:
                self._rng = np.random.RandomState(734421)
        probability = self._rng.random_sample(len(sso_obs[self.app_mag_col]))
        sso_obs["vis"] = np.where(probability <= completeness, 1, 0)
        return sso_obs


class EclStacker(BaseMoStacker):
    """
    Add ecliptic latitude/longitude (ecLat/ecLon) to the slicer ssoObs (in degrees).

    Parameters
    -----------
    ra_col : `str`, optional
        Name of the RA column to convert to ecliptic lat/long. Default 'ra'.
    dec_col : `str`, optional
        Name of the Dec column to convert to ecliptic lat/long. Default 'dec'.
    in_deg : `bool`, optional
        Flag indicating whether RA/Dec are in degrees. Default True.
    """

    cols_added = ["ecLat", "ecLon"]

    def __init__(self, ra_col="ra", dec_col="dec", in_deg=True):
        self.ra_col = ra_col
        self.dec_col = dec_col
        self.in_deg = in_deg
        self.cols_req = [self.ra_col, self.dec_col]
        self.units = ["deg", "deg"]
        self.ecnode = 0.0
        self.ecinc = np.radians(23.439291)

    def _run(self, sso_obs, href, hval):
        ra = sso_obs[self.ra_col]
        dec = sso_obs[self.dec_col]
        if self.in_deg:
            ra = np.radians(ra)
            dec = np.radians(dec)
        x = np.cos(ra) * np.cos(dec)
        y = np.sin(ra) * np.cos(dec)
        z = np.sin(dec)
        xp = x
        yp = np.cos(self.ecinc) * y + np.sin(self.ecinc) * z
        zp = -np.sin(self.ecinc) * y + np.cos(self.ecinc) * z
        sso_obs["ecLat"] = np.degrees(np.arcsin(zp))
        sso_obs["ecLon"] = np.degrees(np.arctan2(yp, xp))
        sso_obs["ecLon"] = sso_obs["ecLon"] % 360
        return sso_obs



__all__ = ("ColInfo",)


from .base_stacker import BaseStacker


class ColInfo:
    """Class to hold the unit and source locations for columns.

    The stacker classes which will generate stacker columns are tracked here, as well as
    some default units for common opsim columns.

    Inspect ColInfo.unit_dict for more information."""

    def __init__(self):
        self.default_data_source = None
        self.default_unit = ""
        self.unit_dict = {
            "fieldId": "#",
            "filter": "filter",
            "seqnNum": "#",
            "expMJD": "MJD",
            "observationStartMJD": "MJD",
            "observationStartLST": "deg",
            "visitExposureTime": "s",
            "slewTime": "s",
            "slewDist": "rad",
            "rotSkyPos": "deg",
            "rotTelPos": "deg",
            "rawSeeing": "arcsec",
            "finSeeing": "arcsec",
            "FWHMeff": "arcsec",
            "FWHMgeom": "arcsec",
            "seeingFwhmEff": "arcsec",
            "seeingFwhmGeom": "arcsec",
            "seeingFwhm500": "arcsec",
            "seeing": "arcsec",
            "airmass": "X",
            "night": "days",
            "moonRA": "rad",
            "moonDec": "rad",
            "moonAlt": "rad",
            "dist2Moon": "rad",
            "filtSkyBrightness": "mag/sq arcsec",
            "skyBrightness": "mag/sq arcsec",
            "fiveSigmaDepth": "mag",
            "solarElong": "degrees",
        }
        # Go through the available stackers and add any units, and identify their
        #   source methods.
        self.source_dict = {}
        for stacker_class in BaseStacker.registry.values():
            stacker = stacker_class()
            for col, units in zip(stacker.cols_added, stacker.units):
                self.source_dict[col] = stacker_class
                self.unit_dict[col] = units
        # Note that a 'unique' list of methods should be built from the resulting returned
        #  methods, at whatever point the derived data columns will be calculated. (i.e. in the driver)

    def get_units(self, col_name):
        """Return the appropriate units for col_name.

        If no units have been defined for a given column, return the default units ('').

        Parameters
        ----------
        col_name : str
            The name of the column

        Returns
        -------
        str
        """
        if col_name is None or col_name not in self.unit_dict:
            return self.default_unit
        else:
            return self.unit_dict[col_name]

    def get_data_source(self, col_name):
        """Identify the appropriate source for a given column.

        For values which are calculated via a stacker, the returned value is the stacker class.
        For values which do not have a recorded source or are known to be coming from the database, the result
        is self.default_data_source (None), which will be assumed to be queryable from the database.

        Parameters
        ----------
        col_name : str
            The name of the column.

        Returns
        -------
        rubin_sim.maf.stacker or None
        """
        if col_name in self.source_dict:
            return self.source_dict[col_name]
        else:
            return self.default_data_source



__all__ = (
    "galplane_nvisits_thresholds",
    "galplane_priority_map_thresholds",
    "GalPlaneFootprintMetric",
    "GalPlaneTimePerFilterMetric",
)

from types import MethodType

import numpy as np

from rubin_sim.maf.maps.galactic_plane_priority_maps import gp_priority_map_components_to_keys

from .base_metric import BaseMetric

# These are a suite of metrics aimed at evaluating high-level quantities regarding galactic plane
# coverage. The metrics here evaluate the coverage (just number of visits and exposure time per filter)
# in relation to the desired coverage from the galactic plane priority map.
# There is a related metric in transientTimeSampling which evaluates the cadence weighted by this same map.

TAU_OBS = np.array([2.0, 5.0, 11.0, 20.0, 46.5, 73.0])


def galplane_nvisits_thresholds(tau_obs, nyears=10):
    """ "Return estimated nvisits required to well-sample lightcurves that need sampling every tau_obs (days).

    This does a very basic estimate, just counting how many visits you would have if you distributed them
    at tau_obs intervals for a period of nyears, assuming a season length of 6.5 years and that visits in
    each night are in pairs.

    Parameters
    ----------
    tau_obs : `np.ndarray` or `list` of `float`
        Timescale that variability must be sampled at, in days.
    nyears : `float`, opt
        Number of years in the survey (as sampled).  Default 10.

    Returns
    -------
    n_visits_thresholds : `np.ndarray`
        Estimated number of visits required to well sample lightcurves which require sampling on tau_obs
    """
    # How many nights in the survey
    nnights_total = 365.25 * nyears
    # Estimate that the observing season is 6.5 months long, not a full year
    # And account for the fact that each 'night' will (currently) need 2 visits
    nnights = nnights_total * (6.5 / 12) * 2
    n_visits_thresholds = nnights / tau_obs
    return n_visits_thresholds


def galplane_priority_map_thresholds(science_map):
    """Return minimum threshold for priority maps, when considering filter balance.

    Parameters
    ----------
    science_map : `str`
        The name of the science map in the galactic plane priority map.

    Returns
    -------
    priority_threshold : `float`
        The minimum threshold to consider from the priority map
    """
    if science_map == "galactic_plane":
        priority_threshold = 0.4
    elif science_map == "combined_map":
        priority_threshold = 0.001
    else:
        priority_threshold = 0.0
    return priority_threshold


# this is a bit of a hack .. it helps us use a variety of tau_obs values,
# and dynamically set up reduce functions
def help_set_reduce_func(obj, metricval, nvisits_thresh):
    def _nvisits_cut(obj, metricval):
        if metricval["n_observations"] >= nvisits_thresh:
            return metricval["map_priority"]
        else:
            return 0

    return _nvisits_cut


class GalPlaneFootprintMetric(BaseMetric):
    """Evaluate the survey overlap with desired regions in the Galactic Plane
    and Magellanic Clouds, by referencing the pre-computed priority maps provided.
    These priority maps are keyed by science area (science_map) and per filter.
    The returned metric values are summed over all filters.

    Parameters
    ----------
    science_map : `str`
        Name of the priority footprint map key to use from the column headers contained in the
        priority_GalPlane_footprint_map_data tables.
    tau_obs : `np.ndarray` or `list` of `float`, opt
        Timescales of minimum-required observations intervals for various classes of time variability.
        Default (None), uses TAU_OBS. In general, this should be left as the default and consistent
        across all galactic-plane oriented metrics.
    mag_cuts : `dict` of `float`, opt
        Magnitudes to use as cutoffs for individual image depths.
        Default None uses a default set of values which correspond roughly to the 50th percentile.
    filter_col : `str`, opt
        Name of the filter column. Default 'filter'.
    m5_col : `str`, opt
        Name of the five-sigma depth column. Default 'fiveSigmaDepth'.
    filterlist : `list` of `str`, opt
        The filters to consider from the priority map and observations.
        Default None uses u, g, r, i, z, and y.
    metricName : `str`, opt
        Name for the metric. Default 'GalPlaneFootprintMetric_{scienceMap}
    """

    def __init__(
        self,
        science_map,
        tau_obs=None,
        mag_cuts=None,
        filter_col="filter",
        m5_col="fiveSigmaDepth",
        filterlist=None,
        **kwargs,
    ):
        self.science_map = science_map
        self.priority_map_threshold = galplane_priority_map_thresholds(self.science_map)
        # tau_obs is an array of minimum-required observation intervals for
        # four categories of time variability
        if tau_obs is not None:
            self.tau_obs = tau_obs
        else:
            self.tau_obs = TAU_OBS
        self.nvisits_threshold = galplane_nvisits_thresholds(self.tau_obs)
        self.filter_col = filter_col
        self.m5_col = m5_col
        if filterlist is not None:
            self.filterlist = filterlist
        else:
            self.filterlist = ["u", "g", "r", "i", "z", "y"]
        if mag_cuts is not None:
            self.mag_cuts = mag_cuts
        else:
            self.mag_cuts = {
                "u": 22.7,
                "g": 24.1,
                "r": 23.7,
                "i": 23.1,
                "z": 22.2,
                "y": 21.4,
            }
        maps = ["GalacticPlanePriorityMap"]
        if "metric_name" not in kwargs:
            metric_name = f"GalplaneFootprintMetric_{self.science_map}"
        else:
            metric_name = kwargs["metric_name"]
            del kwargs["metric_name"]
        for tau, nvisits in zip(self.tau_obs, self.nvisits_threshold):
            tau_reduce_name = f"reduce_Tau_{tau:.1f}".replace(".", "_")
            rfunc = help_set_reduce_func(self, None, nvisits)
            # MethodType(newmethod, self) is how this next line SHOULD go
            # but that doesn't work .. the scope isn't correct somehow.
            # Using this alternate string of tau_reduce_name *does* work.
            setattr(self, tau_reduce_name, MethodType(rfunc, tau_reduce_name))
        super().__init__(
            col=[self.filter_col, self.m5_col],
            metric_name=metric_name,
            maps=maps,
            **kwargs,
        )
        self.reduce_order = {"n_obs": 0, "n_obs_priority": 1}
        for i, tau in enumerate(self.tau_obs):
            r_name = f"Tau_{tau:.1f}".replace(".", "_")
            self.reduce_order[r_name] = i + 2

    def run(self, data_slice, slice_point):
        """Calculate the number of observations that meet the mag_cut values at each slice_point.
        Also calculate the number of observations * the priority map summed over all filter.
        Return both of these values as a dictionary.
        """
        # Check if we want to evaluate this part of the sky, or if the weight is below threshold.
        mapkey = gp_priority_map_components_to_keys("sum", self.science_map)
        priority = slice_point[mapkey]
        if priority <= self.priority_map_threshold:
            return self.badval
        # Count the number of observations per filter, above the mag cuts
        n_obs = 0
        n_obs_priority = 0
        for f in self.filterlist:
            obs_in_filter = np.where(data_slice[self.filter_col] == f)
            above_cut = np.where(data_slice[obs_in_filter][self.m5_col] >= self.mag_cuts[f])
            n_obs += len(above_cut[0])
            mapkey = gp_priority_map_components_to_keys(f, self.science_map)
            n_obs_priority += len(above_cut[0]) * slice_point[mapkey]

        return {
            "n_observations": n_obs,
            "n_obs_priority": n_obs_priority,
            "map_priority": priority,
        }

    def reduce_n_obs(self, metricval):
        return metricval["n_observations"]

    def reduce_n_obs_priority(self, metricval):
        return metricval["n_obs_priority"]


class GalPlaneTimePerFilterMetric(BaseMetric):
    """Evaluate the fraction of exposure time spent in each filter as a fraction of the
    total exposure time dedicated to that healpix in the weighted galactic plane priority maps.

    Parameters
    ----------
    scienceMap : `str`
        Name of the priority footprint map key to use from the column headers contained in the
        priority_GalPlane_footprint_map_data tables.
    magCuts : `dict` of `float`, opt
        Magnitudes to use as cutoffs for individual image depths.
        Default None uses a default set of values which correspond roughly to the 50th percentile.
    mjd_col : `str`, opt
        Name of the observation start MJD column. Default 'observationStartMJD'.
    exp_time_col : `str`, opt
        Name of the exposure time column. Default 'visitExposureTime'.
    filter_col : `str`, opt
        Name of the filter column. Default 'filter'.
    m5_col : `str`, opt
        Name of the five-sigma depth column. Default 'fiveSigmaDepth'.
    filterlist : `list` of `str`, opt
        The filters to consider from the priority map and observations.
        Default None uses u, g, r, i, z, and y.
    metricName : `str`, opt
        Name for the metric. Default 'GalPlaneFootprintMetric_{scienceMap}
    """

    def __init__(
        self,
        science_map,
        mag_cuts=None,
        mjd_col="observationStartMJD",
        exp_time_col="visitExposureTime",
        filter_col="filter",
        m5_col="fiveSigmaDepth",
        filterlist=None,
        **kwargs,
    ):
        self.science_map = science_map
        self.priority_map_threshold = galplane_priority_map_thresholds(self.science_map)
        self.filter_col = filter_col
        self.m5_col = m5_col
        self.mjd_col = mjd_col
        self.exp_time_col = exp_time_col
        if filterlist is not None:
            self.filterlist = filterlist
        else:
            self.filterlist = ["u", "g", "r", "i", "z", "y"]
        if mag_cuts is not None:
            self.mag_cuts = mag_cuts
        else:
            self.mag_cuts = {
                "u": 22.7,
                "g": 24.1,
                "r": 23.7,
                "i": 23.1,
                "z": 22.2,
                "y": 21.4,
            }
        maps = ["GalacticPlanePriorityMap"]
        if "metric_name" not in kwargs:
            metric_name = f"GalplaneTimePerFilter_{self.science_map}"
        else:
            metric_name = kwargs["metric_name"]
            del kwargs["metric_name"]
        super().__init__(
            col=[self.filter_col, self.m5_col, self.mjd_col, self.exp_time_col],
            maps=maps,
            metric_name=metric_name,
            **kwargs,
        )
        # Put the reduce functions into filter order
        for i, f in enumerate(self.filterlist):
            self.reduce_order[f"{f}"] = i

    def run(self, data_slice, slice_point):
        """Calculate the ratio of the actual on-sky exposure time per filter
        compared to the ideal on-sky exposure time per filter at this point on the sky across all filters.
        """
        # Check if we want to evaluate this part of the sky, or if the weight is below threshold.
        weight_all_filters = slice_point[gp_priority_map_components_to_keys("sum", self.science_map)]
        if weight_all_filters <= self.priority_map_threshold:
            return self.badval

        # Calculate the ideal weighting per filter compared to all filters at this point in the sky
        relative_filter_weight = {}
        for f in self.filterlist:
            mapkey = gp_priority_map_components_to_keys(f, self.science_map)
            relative_filter_weight[f] = slice_point[mapkey] / weight_all_filters

        exp_time_per_filter = {}
        for f in self.filterlist:
            # Select observations within the OpSim for the current filter
            # which match the S/N requirement, and extract the exposure times
            # for those observations
            idx1 = np.where(data_slice[self.filter_col] == f)[0]
            idx2 = np.where(data_slice[self.m5_col] >= self.mag_cuts[f])[0]
            match = list(set(idx1).intersection(set(idx2)))

            # Now calculate the actual fraction of exposure time spent
            # in this filter for the current slice_point, relative to the total
            # exposure time spent on this slice_point.
            # Note that this includes dithered observations.
            # If no exposures are expected in this filter, this returns 1
            # on the principle that 100% of the expected observations are
            # provided, and additional data in other filters is usually welcome
            exp_time_per_filter[f] = data_slice[self.exp_time_col][match].sum()

        # Calculate the time on-sky in each filter that overlaps this point, and meets mag_cuts
        total_expt_mag_cut = 0
        for f in self.filterlist:
            total_expt_mag_cut += exp_time_per_filter[f].sum()

        # normalize by the relative filter weight. Ideally metric results are close to 1.
        normalized_exp_time = {}
        for f in self.filterlist:
            if total_expt_mag_cut == 0:
                normalized_exp_time[f] = 0
            else:
                # If no exposures are expected in this filter for this location,
                # thais metric returns the mask val for this filter only.
                if relative_filter_weight[f] > 0:
                    normalized_exp_time[f] = (
                        exp_time_per_filter[f] / total_expt_mag_cut / relative_filter_weight[f]
                    )
                else:
                    normalized_exp_time[f] = self.badval
        return normalized_exp_time

    def reduce_u(self, metricval):
        return metricval["u"]

    def reduce_g(self, metricval):
        return metricval["g"]

    def reduce_r(self, metricval):
        return metricval["r"]

    def reduce_i(self, metricval):
        return metricval["i"]

    def reduce_z(self, metricval):
        return metricval["z"]

    def reduce_y(self, metricval):
        return metricval["y"]



__all__ = (
    "PassMetric",
    "Coaddm5Metric",
    "MaxMetric",
    "AbsMaxMetric",
    "MeanMetric",
    "AbsMeanMetric",
    "MedianMetric",
    "AbsMedianMetric",
    "MinMetric",
    "FullRangeMetric",
    "RmsMetric",
    "RelRmsMetric",
    "SumMetric",
    "CountUniqueMetric",
    "CountMetric",
    "CountRatioMetric",
    "CountSubsetMetric",
    "CountBeyondThreshold",
    "RobustRmsMetric",
    "MaxPercentMetric",
    "AbsMaxPercentMetric",
    "BinaryMetric",
    "FracAboveMetric",
    "FracBelowMetric",
    "PercentileMetric",
    "NoutliersNsigmaMetric",
    "UniqueRatioMetric",
    "MeanAngleMetric",
    "RmsAngleMetric",
    "FullRangeAngleMetric",
    "CountExplimMetric",
    "AngularSpreadMetric",
    "RealMeanMetric",
)

import numpy as np

from .base_metric import BaseMetric

# A collection of commonly used simple metrics, operating on a single column and returning a float.

twopi = 2.0 * np.pi


class PassMetric(BaseMetric):
    """
    Just pass the entire array through
    """

    def __init__(self, cols=None, **kwargs):
        if cols is None:
            cols = []
        super(PassMetric, self).__init__(col=cols, metric_dtype="object", **kwargs)

    def run(self, data_slice, slice_point=None):
        return data_slice


class Coaddm5Metric(BaseMetric):
    """Calculate the coadded m5 value at this gridpoint.

    Parameters
    ----------
    m5Col : `str`, optional
        Name of the m5 column. Default fiveSigmaDepth.
    metric_name : `str`, optional
        Name to associate with the metric output.
    """

    def __init__(self, m5_col="fiveSigmaDepth", metric_name="CoaddM5", **kwargs):
        super(Coaddm5Metric, self).__init__(col=m5_col, metric_name=metric_name, **kwargs)

    def run(self, data_slice, slice_point=None):
        # Running this metric directly from the slicer, this should never come up.
        # However, other metrics call this one and maybe had visits in other filters ..
        if len(data_slice) == 0:
            return self.badval
        return 1.25 * np.log10(np.sum(10.0 ** (0.8 * data_slice[self.colname])))


class MaxMetric(BaseMetric):
    """Calculate the maximum of a simData column slice."""

    def run(self, data_slice, slice_point=None):
        return np.max(data_slice[self.colname])


class AbsMaxMetric(BaseMetric):
    """Calculate the max of the absolute value of a simData column slice."""

    def run(self, data_slice, slice_point=None):
        return np.max(np.abs(data_slice[self.colname]))


class MeanMetric(BaseMetric):
    """Calculate the mean of a simData column slice."""

    def run(self, data_slice, slice_point=None):
        return np.mean(data_slice[self.colname])


class AbsMeanMetric(BaseMetric):
    """Calculate the mean of the absolute value of a simData column slice."""

    def run(self, data_slice, slice_point=None):
        return np.mean(np.abs(data_slice[self.colname]))


class MedianMetric(BaseMetric):
    """Calculate the median of a simData column slice."""

    def run(self, data_slice, slice_point=None):
        return np.median(data_slice[self.colname])


class AbsMedianMetric(BaseMetric):
    """Calculate the median of the absolute value of a simData column slice."""

    def run(self, data_slice, slice_point=None):
        return np.median(np.abs(data_slice[self.colname]))


class MinMetric(BaseMetric):
    """Calculate the minimum of a simData column slice."""

    def run(self, data_slice, slice_point=None):
        return np.min(data_slice[self.colname])


class FullRangeMetric(BaseMetric):
    """Calculate the range of a simData column slice."""

    def run(self, data_slice, slice_point=None):
        return np.max(data_slice[self.colname]) - np.min(data_slice[self.colname])


class RmsMetric(BaseMetric):
    """Calculate the standard deviation of a simData column slice."""

    def run(self, data_slice, slice_point=None):
        return np.std(data_slice[self.colname])


class RelRmsMetric(BaseMetric):
    """Calculate the relative scatter metric (RMS divided by median)."""

    def run(self, data_slice, slice_point=None):
        return np.std(data_slice[self.colname]) / np.median(data_slice[self.colname])


class SumMetric(BaseMetric):
    """Calculate the sum of a simData column slice."""

    def run(self, data_slice, slice_point=None):
        return np.sum(data_slice[self.colname])


class CountUniqueMetric(BaseMetric):
    """Return the number of unique values."""

    def run(self, data_slice, slice_point=None):
        return np.size(np.unique(data_slice[self.colname]))


class UniqueRatioMetric(BaseMetric):
    """Return the number of unique values divided by the total number of values."""

    def run(self, data_slice, slice_point=None):
        ntot = float(np.size(data_slice[self.colname]))
        result = np.size(np.unique(data_slice[self.colname])) / ntot
        return result


class CountMetric(BaseMetric):
    """Count the length of a simData column slice."""

    def __init__(self, col=None, **kwargs):
        super(CountMetric, self).__init__(col=col, **kwargs)
        self.metric_dtype = "int"

    def run(self, data_slice, slice_point=None):
        return len(data_slice[self.colname])


class CountExplimMetric(BaseMetric):
    """Count the number of x second visits.  Useful for rejecting very short exposures
    and counting 60s exposures as 2 visits."""

    def __init__(self, col=None, min_exp=20.0, expected_exp=30.0, exp_col="visitExposureTime", **kwargs):
        self.min_exp = min_exp
        self.expected_exp = expected_exp
        self.exp_col = exp_col
        super().__init__(col=[col, exp_col], **kwargs)
        self.metric_dtype = "int"

    def run(self, data_slice, slice_point=None):
        nv = data_slice[self.exp_col] / self.expected_exp
        nv[np.where(data_slice[self.exp_col] < self.min_exp)[0]] = 0
        nv = np.round(nv)
        return int(np.sum(nv))


class CountRatioMetric(BaseMetric):
    """Count the length of a simData column slice, then divide by 'norm_val'."""

    def __init__(self, col=None, norm_val=1.0, metric_name=None, units="", **kwargs):
        self.norm_val = float(norm_val)
        if metric_name is None:
            metric_name = "CountRatio %s div %.1f" % (col, norm_val)
        super(CountRatioMetric, self).__init__(col=col, metric_name=metric_name, **kwargs)

    def run(self, data_slice, slice_point=None):
        return len(data_slice[self.colname]) / self.norm_val


class CountSubsetMetric(BaseMetric):
    """Count the length of a simData column slice which matches 'subset'."""

    def __init__(self, col=None, subset=None, units="#", **kwargs):
        super(CountSubsetMetric, self).__init__(col=col, units=units, **kwargs)
        self.metric_dtype = "int"
        self.badval = 0
        self.subset = subset

    def run(self, data_slice, slice_point=None):
        count = len(np.where(data_slice[self.colname] == self.subset)[0])
        return count


class CountBeyondThreshold(BaseMetric):
    """Count the number of entries in a data column above or below the threshold."""

    def __init__(self, col=None, lower_threshold=None, upper_threshold=None, **kwargs):
        super().__init__(col=col, **kwargs)
        self.lower_threshold = lower_threshold
        self.upper_threshold = upper_threshold

    def run(self, data_slice, slice_point=None):
        # Look for data values which match the criteria for the thresholds
        if self.upper_threshold is None and self.lower_threshold is None:
            count = len(data_slice)
        elif self.upper_threshold is None:
            count = len(np.where(data_slice[self.colname] > self.lower_threshold)[0])
        elif self.lower_threshold is None:
            count = len(np.where(data_slice[self.colname] < self.upper_threshold)[0])
        else:
            count = len(
                np.where(
                    (data_slice[self.colname] > self.lower_threshold)
                    and (data_slice[self.colname] < self.upper_threshold)
                )[0]
            )
        return count


class RobustRmsMetric(BaseMetric):
    """Use the inter-quartile range of the data to estimate the RMS.
    Robust since this calculation does not include outliers in the distribution.
    """

    def run(self, data_slice, slice_point=None):
        iqr = np.percentile(data_slice[self.colname], 75) - np.percentile(data_slice[self.colname], 25)
        rms = iqr / 1.349  # approximation
        return rms


class MaxPercentMetric(BaseMetric):
    """Return the percent of the data which has the maximum value."""

    def run(self, data_slice, slice_point=None):
        n_max = np.size(np.where(data_slice[self.colname] == np.max(data_slice[self.colname]))[0])
        percent = n_max / float(data_slice[self.colname].size) * 100.0
        return percent


class AbsMaxPercentMetric(BaseMetric):
    """Return the percent of the data which has the absolute value of the max value of the data."""

    def run(self, data_slice, slice_point=None):
        max_val = np.abs(np.max(data_slice[self.colname]))
        n_max = np.size(np.where(np.abs(data_slice[self.colname]) == max_val)[0])
        percent = n_max / float(data_slice[self.colname].size) * 100.0
        return percent


class BinaryMetric(BaseMetric):
    """Return 1 if there is data."""

    def run(self, data_slice, slice_point=None):
        if data_slice.size > 0:
            return 1
        else:
            return self.badval


class FracAboveMetric(BaseMetric):
    """Find the fraction of data values above a given value."""

    def __init__(self, col=None, cutoff=0.5, scale=1, metric_name=None, **kwargs):
        # Col could just get passed in bundle with kwargs, but by explicitly pulling it out
        #  first, we support use cases where class instantiated without explicit 'col=').
        if metric_name is None:
            metric_name = "FracAbove %.2f in %s" % (cutoff, col)
        super(FracAboveMetric, self).__init__(col, metric_name=metric_name, **kwargs)
        self.cutoff = cutoff
        self.scale = scale

    def run(self, data_slice, slice_point=None):
        good = np.where(data_slice[self.colname] >= self.cutoff)[0]
        frac_above = np.size(good) / float(np.size(data_slice[self.colname]))
        frac_above = frac_above * self.scale
        return frac_above


class FracBelowMetric(BaseMetric):
    """Find the fraction of data values below a given value."""

    def __init__(self, col=None, cutoff=0.5, scale=1, metric_name=None, **kwargs):
        if metric_name is None:
            metric_name = "FracBelow %.2f %s" % (cutoff, col)
        super(FracBelowMetric, self).__init__(col, metric_name=metric_name, **kwargs)
        self.cutoff = cutoff
        self.scale = scale

    def run(self, data_slice, slice_point=None):
        good = np.where(data_slice[self.colname] <= self.cutoff)[0]
        frac_below = np.size(good) / float(np.size(data_slice[self.colname]))
        frac_below = frac_below * self.scale
        return frac_below


class PercentileMetric(BaseMetric):
    """Find the value of a column at a given percentile."""

    def __init__(self, col=None, percentile=90, metric_name=None, **kwargs):
        if metric_name is None:
            metric_name = "%.0fth%sile %s" % (percentile, "%", col)
        super(PercentileMetric, self).__init__(col=col, metric_name=metric_name, **kwargs)
        self.percentile = percentile

    def run(self, data_slice, slice_point=None):
        pval = np.percentile(data_slice[self.colname], self.percentile)
        return pval


class NoutliersNsigmaMetric(BaseMetric):
    """Calculate the # of visits less than n_sigma below the mean (n_sigma<0) or
    more than n_sigma above the mean of 'col'.
    """

    def __init__(self, col=None, n_sigma=3.0, metric_name=None, **kwargs):
        self.n_sigma = n_sigma
        self.col = col
        if metric_name is None:
            metric_name = "Noutliers %.1f %s" % (self.n_sigma, self.col)
        super(NoutliersNsigmaMetric, self).__init__(col=col, metric_name=metric_name, **kwargs)
        self.metric_dtype = "int"

    def run(self, data_slice, slice_point=None):
        med = np.mean(data_slice[self.colname])
        std = np.std(data_slice[self.colname])
        boundary = med + self.n_sigma * std
        # If nsigma is positive, look for outliers above median.
        if self.n_sigma >= 0:
            outsiders = np.where(data_slice[self.colname] > boundary)
        # Else look for outliers below median.
        else:
            outsiders = np.where(data_slice[self.colname] < boundary)
        return len(data_slice[self.colname][outsiders])


def _rotate_angles(angles):
    """Private utility for the '*Angle' Metrics below.

    This takes a series of angles between 0-2pi and rotates them so that the
    first angle is at 0, ensuring the biggest 'gap' is at the end of the series.
    This simplifies calculations like the 'mean' and 'rms' or 'fullrange', removing
    the discontinuity at 0/2pi.
    """
    angleidx = np.argsort(angles)
    diffangles = np.diff(angles[angleidx])
    start_to_end = np.array([twopi - angles[angleidx][-1] + angles[angleidx][0]], float)
    if start_to_end < -2.0 * np.pi:
        raise ValueError("Angular metrics expect radians, this seems to be in degrees")
    diffangles = np.concatenate([diffangles, start_to_end])
    maxdiff = np.where(diffangles == diffangles.max())[0]
    if len(maxdiff) > 1:
        maxdiff = maxdiff[-1:]
    if maxdiff == (len(angles) - 1):
        rotation = angles[angleidx][0]
    else:
        rotation = angles[angleidx][maxdiff + 1][0]
    return (rotation, (angles - rotation) % twopi)


class MeanAngleMetric(BaseMetric):
    """Calculate the mean of an angular (degree) simData column slice.

    'MeanAngle' differs from 'Mean' in that it accounts for wraparound at 2pi.
    """

    def run(self, data_slice, slice_point=None):
        """Calculate mean angle via unit vectors.
        If unit vector 'strength' is less than 0.1, then just set mean to 180 degrees
        (as this indicates nearly uniformly distributed angles).
        """
        x = np.cos(np.radians(data_slice[self.colname]))
        y = np.sin(np.radians(data_slice[self.colname]))
        meanx = np.mean(x)
        meany = np.mean(y)
        angle = np.arctan2(meany, meanx)
        radius = np.sqrt(meanx**2 + meany**2)
        mean = angle % twopi
        if radius < 0.1:
            mean = np.pi
        return np.degrees(mean)


class RmsAngleMetric(BaseMetric):
    """Calculate the standard deviation of an angular (degrees) simData column slice.

    'RmsAngle' differs from 'Rms' in that it accounts for wraparound at 2pi.
    """

    def run(self, data_slice, slice_point=None):
        rotation, angles = _rotate_angles(np.radians(data_slice[self.colname]))
        return np.std(np.degrees(angles))


class FullRangeAngleMetric(BaseMetric):
    """Calculate the full range of an angular (degrees) simData column slice.

    'FullRangeAngle' differs from 'FullRange' in that it accounts for wraparound at 2pi.
    """

    def run(self, data_slice, slice_point=None):
        rotation, angles = _rotate_angles(np.radians(data_slice[self.colname]))
        return np.degrees(angles.max() - angles.min())


class AngularSpreadMetric(BaseMetric):
    """Compute the angular spread statistic which measures uniformity of a distribution angles
    accounting for 2pi periodicity.

    The strategy is to first map angles into unit vectors on the unit circle, and then compute the
    2D centroid of those vectors.  A uniform distribution of angles will lead to a distribution of
    unit vectors with mean that approaches the origin.  In contrast, a delta function distribution
    of angles leads to a delta function distribution of unit vectors with a mean that lands on the
    unit circle.

    The angular spread statistic is then defined as 1 - R, where R is the radial offset of the mean
    of the unit vectors derived from the input angles.  R approaches 1 for a uniform distribution
    of angles and 0 for a delta function distribution of angles.

    The optional parameter `period` may be used to specificy periodicity other than 2 pi.
    """

    def __init__(self, col=None, period=2.0 * np.pi, **kwargs):
        # https://en.wikipedia.org/wiki/Directional_statistics#Measures_of_location_and_spread
        # jmeyers314@gmail.com
        self.period = period
        super(AngularSpreadMetric, self).__init__(col=col, **kwargs)

    def run(self, data_slice, slice_point=None):
        # Unit vectors; unwrapped at specified period
        x = np.cos(data_slice[self.colname] * 2.0 * np.pi / self.period)
        y = np.sin(data_slice[self.colname] * 2.0 * np.pi / self.period)
        meanx = np.mean(x)
        meany = np.mean(y)
        # radial offset (i.e., length) of the mean unit vector
        R = np.hypot(meanx, meany)
        return 1.0 - R


class RealMeanMetric(BaseMetric):
    """Calculate the mean of a simData column slice with no nans or infs."""

    def run(self, data_slice, slice_point=None):
        return np.mean(data_slice[self.colname][np.isfinite(data_slice[self.colname])])



__all__ = ("SdssRADecStacker",)


import numpy as np

from .base_stacker import BaseStacker
from .dither_stackers import wrap_ra


class SdssRADecStacker(BaseStacker):
    """convert the p1,p2,p3... columns to radians and wrap them"""

    cols_added = ["RA1", "Dec1", "RA2", "Dec2", "RA3", "Dec3", "RA4", "Dec4"]

    def __init__(self, pcols=["p1", "p2", "p3", "p4", "p5", "p6", "p7", "p8"]):
        """The p1,p2 columns represent the corners of chips.  Could generalize this a bit."""
        self.units = ["rad"] * 8
        self.cols_req = pcols

    def _run(self, sim_data, cols_present=False):
        if cols_present:
            # Assume this is unusual enough to run that you really mean it.
            pass
        for pcol, newcol in zip(self.cols_req, self.cols_added):
            if newcol[0:2] == "RA":
                sim_data[newcol] = wrap_ra(np.radians(sim_data[pcol]))
            else:
                sim_data[newcol] = np.radians(sim_data[pcol])
        return sim_data



__all__ = ("PrestoColorKNePopMetric", "generate_presto_pop_slicer")

import os
import pickle
import warnings
from itertools import combinations

import numpy as np
import pandas as pd
from rubin_scheduler.data import get_data_dir
from rubin_scheduler.utils import survey_start_mjd, uniform_sphere

import rubin_sim.maf.metrics as metrics
import rubin_sim.maf.slicers as slicers
from rubin_sim.phot_utils import DustValues

from .kne_metrics import KnLc


def radec2gal(ra, dec):
    """convert from ra/dec to galactic l/b"""
    from astropy import units as u
    from astropy.coordinates import SkyCoord

    c = SkyCoord(ra=ra, dec=dec, unit=(u.degree, u.degree))

    gal_l = c.galactic.l.degree
    gal_b = c.galactic.b.degree
    return gal_l, gal_b


def _load_hash(
    file_galactic="TotalCubeNorm_1000Obj.pkl",
    file_extragalactic="TotalCubeNorm_1000Obj.pkl",
    skyregion="extragalactic",
):
    """Helper function to load large hash table.

    Because this is kept outside the metric attributes, it allows easy resuse.
    Note that this does mean running different sky regions at the same
    time may end up thrashing the data in the hash/hash table.

    Parameters
    ----------
    skyregion : `str`
        The skyregion of interst.
        Only two options: 'galactic' and 'extragalactic'
    filePathGalactic : `str`
        File containing galactic Presto-Color phase space information
    filePathExtragalactic : `str`
        File containing galactic Presto-Color phase space information
    """

    if hasattr(_load_hash, "InfoDict"):
        if skyregion == _load_hash.skyregion:
            return _load_hash.InfoDict, _load_hash.HashTable

    data_dir = get_data_dir()
    if skyregion == "galactic":
        file_path = os.path.join(data_dir, "maf", file_galactic)
    elif skyregion == "extragalactic":
        file_path = os.path.join(data_dir, "maf", file_extragalactic)

    with open(file_path, "rb") as f:
        _load_hash.InfoDict = pickle.load(f)
        _load_hash.HashTable = pickle.load(f)
    _load_hash.skyregion = skyregion
    return _load_hash.InfoDict, _load_hash.HashTable


def generate_presto_pop_slicer(
    skyregion="galactic",
    t_start=1,
    t_end=3652,
    n_events=10000,
    seed=42,
    n_files=100,
    d_min=10,
    d_max=300,
    gb_cut=20,
):
    """Generate a population of KNe events, and put the info about them
    into a UserPointSlicer object
    Parameters
    ----------
    skyregion : `str`
        The skyregion of interest.
        Only two options: 'galactic' and 'extragalactic'
    t_start : `float`
        The night to start kilonova events on (days)
    t_end : `float`
        The final night of kilonova events
    n_events : `int`
        The number of kilonova events to generate
    seed : `float`
        The seed passed to np.random
    n_files : `int`
        The number of different kilonova lightcurves to use
    d_min : `float` or `int`
        Minimum luminosity distance (Mpc)
    d_max : `float` or `int`
        Maximum luminosity distance (Mpc)
    """

    def rndm(a, b, g, size=1):
        """Power-law gen for pdf(x) proportional to x^{g-1} for a<=x<=b"""
        r = np.random.random(size=size)
        ag, bg = a**g, b**g
        return (ag + (bg - ag) * r) ** (1.0 / g)

    ra, dec = uniform_sphere(n_events, seed=seed)

    # Convert ra, dec to gl, gb
    gl, gb = radec2gal(ra, dec)

    # Determine if the object is in the Galaxy plane
    if skyregion == "galactic":  # keep the galactic events
        ra = ra[np.abs(gb) < gb_cut]
        dec = dec[np.abs(gb) < gb_cut]
    elif skyregion == "extragalactic":  # keep the extragalactic events.
        ra = ra[np.abs(gb) > gb_cut]
        dec = dec[np.abs(gb) > gb_cut]
    else:
        warnings.warn("Skyregion %s not recognized, using whole sky" % skyregion)

    n_events = len(ra)

    peak_times = np.random.uniform(low=t_start, high=t_end, size=n_events)
    file_indx = np.floor(np.random.uniform(low=0, high=n_files, size=n_events)).astype(int)

    # Define the distance
    distance = rndm(d_min, d_max, 4, size=n_events)

    # Set up the slicer to evaluate the catalog we just made
    slicer = slicers.UserPointsSlicer(ra, dec, lat_lon_deg=True, badval=0)
    # Add any additional information about each object to the slicer
    slicer.slice_points["peak_time"] = peak_times
    slicer.slice_points["file_indx"] = file_indx
    slicer.slice_points["distance"] = distance
    return slicer


class PrestoColorKNePopMetric(metrics.BaseMetric):
    def __init__(
        self,
        metric_name="KNePopMetric",
        mjd_col="observationStartMJD",
        m5_col="fiveSigmaDepth",
        filter_col="filter",
        night_col="night",
        pts_needed=2,
        file_list=None,
        mjd0=None,
        output_lc=False,
        skyregion="galactic",
        thr=0.003,
        **kwargs,
    ):
        """
        Parameters
        ----------
        file_list : `str` or None, optional
            File containing input lightcurves
        mjd0 : `float`, optional
            MJD of the start of the survey.
        output_lc : `bool`, optional
            Flag to whether or not to output lightcurve for each object.
        skyregion : `str`, optional
            The skyregion of interest.
            Only two options: 'galactic' and 'extragalactic'
        thr : `float`, optional
            Threshold for "classification" of events via the Score_S
        """
        maps = ["DustMap"]
        self.mjd_col = mjd_col
        self.m5_col = m5_col
        self.filter_col = filter_col
        self.night_col = night_col
        # Boolean variable, if True the light curve will be exported
        self.output_lc = output_lc
        self.thr = thr
        self.skyregion = skyregion
        # read in file as light curve object;
        self.lightcurves = KnLc(file_list=file_list)
        self.mjd0 = survey_start_mjd() if mjd0 is None else mjd0

        dust_properties = DustValues()
        self.ax1 = dust_properties.ax1

        cols = [self.mjd_col, self.m5_col, self.filter_col, self.night_col]
        super().__init__(col=cols, units="Detected, 0 or 1", metric_name=metric_name, maps=maps, **kwargs)

    def _presto_color_detect(self, around_peak, filters):
        """Detection criteria of presto cadence:
        at least three detections at two filters;

        Parameters
        ----------
        around_peak : `np.ndarray`, (N,)
            indexes corresponding to 5sigma detections
        filters : `np.ndarray`, (N,)
            filters in which detections happened
        """
        result = 1

        if np.size(around_peak) < 3:
            result = 0

        flts, flts_count = np.unique(
            filters,
            return_counts=True,
        )
        if np.size(flts) < 2:
            result = 0
        elif np.max(flts_count) < 2:
            # if no filters have visits larger than 2, set detection false
            result = 0

        return result

    def _enquiry(self, hash_table, info_dict, band1, band2, d_t1, d_t2, d_mag, color):
        """
        Return the value in the probability cube provided the coordinates
        in the Presto-color phase space of an observation triplet.

        Parameters
        ----------
        hash_table : `np.ndarray`, (N,)
            Contains the values of the 6-D Presto-color phase space
        info_dict : `dict`
            Contains the essential information of the hash_table abobe.
        band1, band2 : `str`, `str`
            The two filters that comprise the Presto-color observation triplet.
            The filters are the 6 bands of LSST: u, g, r, i, z, y.
            Band1 and band2 should be different.
        d_t1, d_t2 : `float`, `float`
            The time gaps of the Presto-color observation triplet.
        d_mag : `float`
            The magnitude change between from the observations of the same band
        color : `float`
            The difference in magnitude of observations in different bands.

        hash_table and info_dict have to be loaded from premade data
        Presto-color data file.
        """

        #         if abs(d_t1) > abs(d_t1-d_t2):
        #             d_t1, d_t2 = d_t1-d_t2, -d_t2

        if not (
            info_dict["BinMag"][0] <= d_mag < info_dict["BinMag"][-1]
            and info_dict["BinColor"][0] <= color < info_dict["BinColor"][-1]
        ):
            return 0

        ind1 = info_dict["BandPairs"].index(band1 + band2)

        time_pair_grid = [
            info_dict["dT1s"][abs(d_t1 - info_dict["dT1s"]).argmin()],
            info_dict["dT2s"][abs(d_t2 - info_dict["dT2s"]).argmin()],
        ]

        ind2 = np.where((info_dict["TimePairs"] == time_pair_grid).all(axis=1))[0][0]
        ind3 = np.where(d_mag >= info_dict["BinMag"])[0][-1]
        ind4 = np.where(color >= info_dict["BinColor"])[0][-1]

        return hash_table[ind1, ind2, ind3, ind4]

    def _get_score(self, result, hash_table, info_dict, thr):
        """Get the score of a strategy from the Presto-color perspective.

        Parameters
        ----------
        result : `pd.DataFrame`
            Dataframe that contains the results of the observations.
            The columns include
            t: the time of the observation
            mag: the detected magnitude
            maglim: the limit fiveSigmaDepth that can be detected
            filter: the filter used for the observation
        hash_table : `np.ndarray`, (N,)
            Contains the values of the 6-D Presto-color phase space
        info_dict : `dict`
            Contains the essential information of the hash_table abobe.
        scoreType : `str`
            Two types of scores were designed:
            'S' type involves a threshold,
            'P' type work without a threshold.
        thr : `float`
            The threashold need for type 'S' score.
            The default value is 0.003 (3-sigma)

        hash_table and info_dict have to be loaded from the premade
        Presto-color data file.
        """

        time_lim1 = 8.125 / 24  # 8 h 7.5 min
        time_lim2 = 32.25 / 24  # 32 h 15 min

        detects = result[result.mag < result.maglim]

        # reset index
        detects = detects.reset_index(drop=True)

        # Times for valid detections
        ts = detects.t.values
        # Find out the differences between each pair
        d_ts = ts.reshape(1, len(ts)) - ts.reshape(len(ts), 1)

        # The time differences should be within 32 hours (2 nights)
        d_tindex0, d_tindex1 = np.where(abs(d_ts) < time_lim2)

        phase_space_coords = []

        # loop through the rows of the matrix of valid time differences
        for ii in range(d_ts.shape[0]):
            groups_of_three = np.array(
                [
                    [ii] + list(jj)
                    for jj in list(combinations(d_tindex1[(d_tindex0 == ii) * (d_tindex1 > ii)], 2))
                ]
            )

            for indices in groups_of_three:
                bands = detects["filter"][indices].values

                # print('bands: ', bands)
                if len(np.unique(bands)) != 2:
                    continue

                # The band appears once will be band2
                occurence = np.array([np.count_nonzero(ii == bands) for ii in bands])
                # The index of observation in band2
                index2 = indices[occurence == 1][0]
                # The index of the first observation in band1
                index11 = indices[occurence == 2][0]
                # The index of the second observation in band1
                index12 = indices[occurence == 2][1]

                if (
                    abs(d_ts[index12, index2]) < abs(d_ts[index11, index2])
                    and abs(d_ts[index12, index2]) < time_lim1
                ):
                    index11, index12 = index12, index11
                elif abs(d_ts[index11, index2]) > time_lim1:
                    continue

                d_t1 = d_ts[index11, index2]
                d_t2 = d_ts[index11, index12]

                band1 = bands[occurence == 2][0]
                band2 = bands[occurence == 1][0]

                if band1 + band2 == "uy" or band1 + band2 == "yu":
                    continue

                d_mag = (detects.mag[index11] - detects.mag[index12]) * np.sign(d_t2)
                color = detects.mag[index11] - detects.mag[index2]

                phase_space_coords.append([band1, band2, d_t1, d_t2, d_mag, color])

        score_s = 0
        score_p = [0]

        for phase_space_coord in phase_space_coords:
            rate = self._enquiry(hash_table, info_dict, *phase_space_coord)

            if score_s == 0 and rate < thr:
                score_s = 1

            score_p.append((1 - rate))

        return score_s, max(score_p)

    def run(self, data_slice, slice_point=None):
        data_slice.sort(order=self.mjd_col)
        result = {}
        t = data_slice[self.mjd_col] - self.mjd0 - slice_point["peak_time"]
        mags = np.zeros(t.size, dtype=float)

        for filtername in np.unique(data_slice[self.filter_col]):
            infilt = np.where(data_slice[self.filter_col] == filtername)
            mags[infilt] = self.lightcurves.interp(t[infilt], filtername, lc_indx=slice_point["file_indx"])
            # Apply dust extinction on the light curve
            a_x = self.ax1[filtername] * slice_point["ebv"]
            mags[infilt] += a_x

            distmod = 5 * np.log10(slice_point["distance"] * 1e6) - 5.0
            mags[infilt] += distmod

        # Find the detected points
        around_peak = np.where((t > 0) & (t < 30) & (mags < data_slice[self.m5_col]))[0]
        # Filters in which the detections happened
        filters = data_slice[self.filter_col][around_peak]

        # presto color
        result["presto_color_detect"] = self._presto_color_detect(around_peak, filters)

        # Export the light curve
        idx = np.where(mags < 100)[0]
        lc = {
            "t": data_slice[self.mjd_col][idx],
            "mag": mags[idx],
            "maglim": data_slice[self.m5_col][idx],
            "filter": data_slice[self.filter_col][idx],
        }

        if self.output_lc is True:
            result["lc"] = lc
            result["slice_point"] = slice_point

        if result["presto_color_detect"] == 1:
            info_dict, hash_table = _load_hash(skyregion=self.skyregion)
            result["scoreS"], result["scoreP"] = self._get_score(
                pd.DataFrame(lc),
                hash_table=hash_table,
                info_dict=info_dict,
                thr=self.thr,
            )
        else:
            result["scoreS"] = 0
            result["scoreP"] = 0
        return result

    def reduce_presto_color_detect(self, metric):
        return metric["presto_color_detect"]

    def reduce_score_s(self, metric):
        return metric["scoreS"]

    def reduce_score_p(self, metric):
        return metric["scoreP"]



__all__ = ("SurfaceBrightLimitMetric",)

import warnings

import numpy as np

from rubin_sim.maf.utils import load_inst_zeropoints

from .base_metric import BaseMetric


def surface_brightness_limit_approx(
    zp,
    k,
    airmass,
    mu_sky,
    rn=8.8,
    pixscale=0.2,
    nsigma=3.0,
    t_exp=30.0,
    tot_area=100.0,
    mag_diff_warn=0.1,
):
    """Compute surface brightness limit in 3 limiting cases, return the brightest.

    Algerbra worked out in this technote:
    https://github.com/lsst-sims/smtn-016

    Parameters
    ---------
    zp : `float`
        Telescope zeropoint (mags)
    k : `float`
        Atmospheric extinction term
    airmass : `float`
        Airmass
    mu_sky : `float`
        Surface brightness of the sky (mag/sq arcsec)
    rn : `float` (8.8)
        Readnoise in electrons
    pixscale : `float` (0.2)
        Arcseconds per pixel
    nsigma : `float` (3)
        The SNR to demand
    t_exp : `float` (30.)
        Exposure time (seconds)
    tot_area : `float` (100)
        The total area measuring over (sq arcsec)
    mag_diff_warn : `float` (0.1)
        If the limiting cases are within mag_diff_warn, throw a warning
        that the surface brightness limit may be an overestimate.

    Returns
    -------
    surface brightness limit in mags/sq arcsec
    aka the surface brightness that reaches SNR=nsigma when measured over tot_area.
    """

    a_pix = pixscale**2

    n_pix = tot_area / a_pix

    # Sky limited case
    mu_sky_lim = (
        -1.25 * np.log10(nsigma**2 / (a_pix * t_exp * n_pix)) + 0.5 * mu_sky + 0.5 * zp - k * airmass
    )

    # Source limited case
    # XXX--double check this algerbra. Pretty sure it's right now.
    mu_source_lim = -5 * np.log10(nsigma) + 2.5 * np.log10(n_pix * a_pix * t_exp) + zp - k * airmass

    # Readnoise limited case
    mu_rn_lim = -2.5 * np.log10(nsigma * rn / (t_exp * a_pix * n_pix**0.5)) + zp - k * airmass

    d1 = np.min(np.abs(mu_sky_lim - mu_source_lim))
    d2 = np.min(np.abs(mu_sky_lim - mu_rn_lim))
    d3 = np.min(np.abs(mu_rn_lim - mu_source_lim))

    if np.min([d1, d2, d3]) < mag_diff_warn:
        warnings.warn(
            "Limiting magnitudes in different cases are within %.3f mags, result may be too optimistic by up 0.38 mags/sq arcsec."
            % mag_diff_warn
        )

    result = np.vstack([mu_sky_lim, mu_source_lim, mu_rn_lim])
    result = np.min(result, axis=0)

    return result


class SurfaceBrightLimitMetric(BaseMetric):
    """Gaussian limit, ignoring systematic errors in photometry

    Parameters
    ----------
    pixscale : `float` (0.2)
        Pixelscale, Arcseconds per pixel
    nsigma : `float` (3)
        The detection limit (usuall 3 or 5)
    tot_area : `float` (100)
        Total sky area summed over, square arcseconds
    zpt : `dict` of `float` (None)
        telescope zeropoints. If None, computed from phot_utils
    k_atm : `dict` of `float` (None)
        Atmospheric extinction parameters. If None, computed from phot_utils
    readnoise : `float` (8.8)
        Readnoise in electrons
    """

    def __init__(
        self,
        pixscale=0.2,
        nsigma=3.0,
        tot_area=100.0,
        filter_col="filter",
        units="mag/sq arcsec",
        airmass_col="airmass",
        exptime_col="visitExposureTime",
        metric_name="SurfaceBrightLimit",
        skybrightness_col="skyBrightness",
        nexp_col="numExposures",
        zpt=None,
        k_atm=None,
        readnoise=8.8,
        **kwargs,
    ):
        super().__init__(
            col=[filter_col, airmass_col, exptime_col, skybrightness_col, nexp_col],
            units=units,
            metric_name=metric_name,
            **kwargs,
        )
        self.filter_col = filter_col
        self.airmass_col = airmass_col
        self.exptime_col = exptime_col
        self.skybrightness_col = skybrightness_col
        self.nexp_col = nexp_col

        self.readnoise = readnoise
        self.pixscale = pixscale
        self.nsigma = nsigma
        self.tot_area = tot_area

        # Compute default zeropoints
        if zpt is None:
            zp_inst, k_atm = load_inst_zeropoints()
            self.zpt = zp_inst
            self.k_atm = k_atm
        else:
            self.zpt = zpt
            self.k_atm = k_atm

    def run(self, data_slice, slice_point):
        filtername = np.unique(data_slice[self.filter_col])
        if np.size(filtername) > 1:
            ValueError("Can only coadd depth in single filter, got filters %s" % filtername)
        filtername = filtername[0]

        # Scale up readnoise if the visit was split into multiple snaps
        readnoise = self.readnoise * np.sqrt(data_slice[self.nexp_col])

        sb_per_visit = surface_brightness_limit_approx(
            self.zpt[filtername],
            self.k_atm[filtername],
            data_slice[self.airmass_col],
            data_slice[self.skybrightness_col],
            rn=readnoise,
            pixscale=self.pixscale,
            nsigma=self.nsigma,
            t_exp=data_slice[self.exptime_col],
            tot_area=self.tot_area,
        )

        coadd_sb_limit = 1.25 * np.log10(np.sum(10 ** (0.8 * sb_per_visit)))
        return coadd_sb_limit



__all__ = ("HealpixComCamSlicer",)

import warnings
from functools import wraps

import matplotlib.path as mplPath
import numpy as np
import rubin_scheduler.utils as simsUtils

from .healpix_slicer import HealpixSlicer

# The names of the chips in the central raft, aka, ComCam
center_raft_chips = [
    "R:2,2 S:0,0",
    "R:2,2 S:0,1",
    "R:2,2 S:0,2",
    "R:2,2 S:1,0",
    "R:2,2 S:1,1",
    "R:2,2 S:1,2",
    "R:2,2 S:2,0",
    "R:2,2 S:2,1",
    "R:2,2 S:2,2",
]


class HealpixComCamSlicer(HealpixSlicer):
    """Slicer that uses the ComCam footprint to decide if observations overlap a healpixel center"""

    def __init__(
        self,
        nside=128,
        lon_col="fieldRA",
        lat_col="fieldDec",
        lat_lon_deg=True,
        verbose=True,
        badval=np.nan,
        use_cache=True,
        leafsize=100,
        use_camera=False,
        rot_sky_pos_col_name="rotSkyPos",
        mjdColName="observationStartMJD",
        chipNames=center_raft_chips,
        side_length=0.7,
    ):
        """
        Parameters
        ----------
        side_length : float (0.7)
            How large is a side of the raft (degrees)
        """
        radius = side_length / 2.0 * np.sqrt(2.0)
        super(HealpixComCamSlicer, self).__init__(
            nside=nside,
            lon_col=lon_col,
            lat_col=lat_col,
            lat_lon_deg=lat_lon_deg,
            verbose=verbose,
            badval=badval,
            use_cache=use_cache,
            leafsize=leafsize,
            radius=radius,
            use_camera=use_camera,
            rot_sky_pos_col_name=rot_sky_pos_col_name,
            mjdColName=mjdColName,
            chipNames=chipNames,
        )
        self.side_length = np.radians(side_length)
        self.corners_x = np.array(
            [
                -self.side_length / 2.0,
                -self.side_length / 2.0,
                self.side_length / 2.0,
                self.side_length / 2.0,
            ]
        )
        self.corners_y = np.array(
            [
                self.side_length / 2.0,
                -self.side_length / 2.0,
                -self.side_length / 2.0,
                self.side_length / 2.0,
            ]
        )
        # Need the rotation even if not using the camera
        self.columns_needed.append(rot_sky_pos_col_name)
        self.columns_needed = list(set(self.columns_needed))

        # The 3D search radius for things inside the raft
        self.side_radius = simsUtils.xyz_angular_radius(side_length / 2.0)

    def setup_slicer(self, sim_data, maps=None):
        """Use sim_data[self.lon_col] and sim_data[self.lat_col] (in radians) to set up KDTree.

        Parameters
        -----------
        sim_data : numpy.recarray
            The simulated data, including the location of each pointing.
        maps : list of rubin_sim.maf.maps objects, optional
            List of maps (such as dust extinction) that will run to build up additional metadata at each
            slice_point. This additional metadata is available to metrics via the slice_point dictionary.
            Default None.
        """
        if maps is not None:
            if self.cache_size != 0 and len(maps) > 0:
                warnings.warn(
                    "Warning:  Loading maps but cache on." "Should probably set use_cache=False in slicer."
                )
            self._run_maps(maps)
        self._setRad(self.radius)
        if self.use_camera:
            self._setupLSSTCamera()
            self._presliceFootprint(sim_data)
        else:
            if self.latLonDeg:
                self._build_tree(
                    np.radians(sim_data[self.lon_col]),
                    np.radians(sim_data[self.lat_col]),
                    self.leafsize,
                )
            else:
                self._build_tree(sim_data[self.lon_col], sim_data[self.lat_col], self.leafsize)

        @wraps(self._slice_sim_data)
        def _slice_sim_data(islice):
            """Return indexes for relevant opsim data at slice_point
            (slice_point=lon_col/lat_col value .. usually ra/dec)."""

            # Build dict for slice_point info
            slice_point = {"sid": islice}
            if self.use_camera:
                indices = self.sliceLookup[islice]
                slice_point["chipNames"] = self.chipNames[islice]
            else:
                sx, sy, sz = simsUtils._xyz_from_ra_dec(
                    self.slice_points["ra"][islice], self.slice_points["dec"][islice]
                )
                # Anything within half the side length is good no matter what rotation angle
                # the camera is at
                indices = self.opsimtree.query_ball_point((sx, sy, sz), self.side_radius)
                # Now the larger radius. Need to make it an array for easy subscripting
                initial_indices = np.array(self.opsimtree.query_ball_point((sx, sy, sz), self.rad), dtype=int)
                # remove the indices we already know about
                initial_indices = initial_indices[np.in1d(initial_indices, indices, invert=True)]

                if self.latLonDeg:
                    lat = np.radians(sim_data[self.lat_col][initial_indices])
                    lon = np.radians(sim_data[self.lon_col][initial_indices])
                    cos_rot = np.cos(np.radians(sim_data[self.rotSkyPosColName][initial_indices]))
                    sin_rot = np.cos(np.radians(sim_data[self.rotSkyPosColName][initial_indices]))
                else:
                    lat = sim_data[self.lat_col][initial_indices]
                    lon = sim_data[self.lon_col][initial_indices]
                    cos_rot = np.cos(sim_data[self.rotSkyPosColName][initial_indices])
                    sin_rot = np.sin(sim_data[self.rotSkyPosColName][initial_indices])
                # loop over the observations that might be overlapping the healpix, check each
                for i, ind in enumerate(initial_indices):
                    # Rotate the camera
                    x_rotated = self.corners_x * cos_rot[i] - self.corners_y * sin_rot[i]
                    y_rotated = self.corners_x * sin_rot[i] + self.corners_y * cos_rot[i]
                    # How far is the pointing center from the healpix center
                    xshift, yshift = simsUtils.gnomonic_project_toxy(
                        lon[i],
                        lat[i],
                        self.slice_points["ra"][islice],
                        self.slice_points["dec"][islice],
                    )

                    x_rotated += xshift
                    y_rotated += yshift
                    # Use matplotlib to make a polygon
                    bbPath = mplPath.Path(
                        np.array(
                            [
                                [x_rotated[0], y_rotated[0]],
                                [x_rotated[1], y_rotated[1]],
                                [x_rotated[2], y_rotated[2]],
                                [x_rotated[3], y_rotated[3]],
                                [x_rotated[0], y_rotated[0]],
                            ]
                        )
                    )
                    # Check if the slice_point is inside the image corners and append to list
                    if bbPath.contains_point((0.0, 0.0)):
                        indices.append(ind)

            # Loop through all the slice_point keys. If the first dimension of slice_point[key] has
            # the same shape as the slicer, assume it is information per slice_point.
            # Otherwise, pass the whole slice_point[key] information. Useful for stellar LF maps
            # where we want to pass only the relevant LF and the bins that go with it.

            for key in self.slice_points:
                if len(np.shape(self.slice_points[key])) == 0:
                    keyShape = 0
                else:
                    keyShape = np.shape(self.slice_points[key])[0]
                if keyShape == self.nslice:
                    slice_point[key] = self.slice_points[key][islice]
                else:
                    slice_point[key] = self.slice_points[key]

            return {"idxs": indices, "slice_point": slice_point}

        setattr(self, "_slice_sim_data", _slice_sim_data)



__all__ = ("OneDBinnedData",)


import matplotlib.pyplot as plt
import numpy as np

from rubin_sim.maf.utils import percentile_clipping

from .plot_handler import BasePlotter


class OneDBinnedData(BasePlotter):
    def __init__(self):
        self.plot_type = "BinnedData"
        self.object_plotter = False
        self.default_plot_dict = {
            "title": None,
            "label": None,
            "xlabel": None,
            "ylabel": None,
            "filled": False,
            "alpha": 0.5,
            "linestyle": "-",
            "linewidth": 1,
            "log_scale": False,
            "percentile_clip": None,
            "x_min": None,
            "x_max": None,
            "y_min": None,
            "y_max": None,
            "fontsize": None,
            "figsize": None,
            "grid": True,
        }

    def __call__(self, metric_values, slicer, user_plot_dict, fignum=None):
        """
        Plot a set of oneD binned metric data.
        """
        if slicer.slicer_name != "OneDSlicer":
            raise ValueError("OneDBinnedData plotter is for use with OneDSlicer")
        if "bins" not in slicer.slice_points:
            err_message = 'OneDSlicer must contain "bins" in slice_points metadata.'
            err_message += " SlicePoints only contains keys %s." % (slicer.slice_points.keys())
            raise ValueError(err_message)
        plot_dict = {}
        plot_dict.update(self.default_plot_dict)
        plot_dict.update(user_plot_dict)
        fig = plt.figure(fignum, figsize=plot_dict["figsize"])
        # Plot the histogrammed data.
        leftedge = slicer.slice_points["bins"][:-1]
        width = np.diff(slicer.slice_points["bins"])
        if plot_dict["filled"]:
            plt.bar(
                leftedge,
                metric_values.filled(),
                width,
                label=plot_dict["label"],
                linewidth=0,
                alpha=plot_dict["alpha"],
                log=plot_dict["log_scale"],
                color=plot_dict["color"],
            )
        else:
            good = np.where(metric_values.mask == False)
            x = np.ravel(list(zip(leftedge[good], leftedge[good] + width[good])))
            y = np.ravel(list(zip(metric_values[good], metric_values[good])))
            if plot_dict["log_scale"]:
                plt.semilogy(
                    x,
                    y,
                    label=plot_dict["label"],
                    color=plot_dict["color"],
                    linestyle=plot_dict["linestyle"],
                    linewidth=plot_dict["linewidth"],
                    alpha=plot_dict["alpha"],
                )
            else:
                plt.plot(
                    x,
                    y,
                    label=plot_dict["label"],
                    color=plot_dict["color"],
                    linestyle=plot_dict["linestyle"],
                    linewidth=plot_dict["linewidth"],
                    alpha=plot_dict["alpha"],
                )
        if "ylabel" in plot_dict:
            plt.ylabel(plot_dict["ylabel"], fontsize=plot_dict["fontsize"])
        if "xlabel" in plot_dict:
            plt.xlabel(plot_dict["xlabel"], fontsize=plot_dict["fontsize"])
        # Set y limits (either from values in args, percentile_clipping or compressed data values).
        if plot_dict["percentile_clip"] is not None:
            y_min, y_max = percentile_clipping(
                metric_values.compressed(), percentile=plot_dict["percentile_clip"]
            )
            if plot_dict["y_min"] is None:
                plot_dict["y_min"] = y_min
            if plot_dict["y_max"] is None:
                plot_dict["y_max"] = y_max

        if plot_dict["grid"]:
            plt.grid(plot_dict["grid"], alpha=0.3)

        if plot_dict["y_min"] is None and metric_values.filled().min() == 0:
            plot_dict["y_min"] = 0

        # Set y and x limits, if provided.
        if plot_dict["y_min"] is not None:
            plt.ylim(bottom=plot_dict["y_min"])
        if plot_dict["y_max"] is not None:
            plt.ylim(top=plot_dict["y_max"])
        if plot_dict["x_min"] is not None:
            plt.xlim(left=plot_dict["x_min"])
        if plot_dict["x_max"] is not None:
            plt.xlim(right=plot_dict["x_max"])
        plt.title(plot_dict["title"])
        return fig.number



__all__ = ("SNCadenceMetric",)

import numpy as np

import rubin_sim.maf.metrics as metrics


class SNCadenceMetric(metrics.BaseMetric):
    """
    Metric to estimate the redshift limit for faint supernovae (x1,color) = (-2.0,0.2)

    Parameters
    ----------
    list : str, optional
        Name of the columns used to estimate the metric
    coadd :  bool, optional
        to make "coaddition" per night (uses snStacker)
        Default True
    lim_sn : class, optional
       Reference data used to estimate redshift values (interpolation)
    """

    def __init__(
        self,
        metric_name="SNCadenceMetric",
        mjd_col="observationStartMJD",
        ra_col="fieldRA",
        dec_col="fieldDec",
        filter_col="filter",
        m5_col="fiveSigmaDepth",
        exptime_col="visitExposureTime",
        night_col="night",
        obsid_col="observationId",
        nexp_col="numExposures",
        vistime_col="visitTime",
        coadd=True,
        lim_sn=None,
        **kwargs,
    ):
        self.mjd_col = mjd_col
        self.m5_col = m5_col
        self.filter_col = filter_col
        self.ra_col = ra_col
        self.dec_col = dec_col
        self.exptime_col = exptime_col
        self.season_col = "season"
        self.night_col = night_col
        self.obsid_col = obsid_col
        self.nexp_col = nexp_col
        self.vistime_col = vistime_col

        cols = [
            self.night_col,
            self.m5_col,
            self.filter_col,
            self.mjd_col,
            self.obsid_col,
            self.nexp_col,
            self.vistime_col,
            self.exptime_col,
            self.season_col,
        ]
        if coadd:
            cols += ["coadd"]

        super(SNCadenceMetric, self).__init__(col=cols, metric_name=metric_name, **kwargs)

        self.filter_names = np.array(["u", "g", "r", "i", "z", "y"])

        self.lim_sn = lim_sn

    def run(self, data_slice, slice_point=None):
        # Cut down to only include filters in correct wave range.

        good_filters = np.in1d(data_slice["filter"], self.filter_names)
        data_slice = data_slice[good_filters]
        if data_slice.size == 0:
            return None
        data_slice.sort(order=self.mjd_col)

        r = []
        field_ra = np.mean(data_slice[self.ra_col])
        field_dec = np.mean(data_slice[self.dec_col])
        band = np.unique(data_slice[self.filter_col])[0]

        sel = data_slice
        bins = np.arange(np.floor(sel[self.mjd_col].min()), np.ceil(sel[self.mjd_col].max()), 1.0)
        c, b = np.histogram(sel[self.mjd_col], bins=bins)
        if (c.mean() < 1.0e-8) | np.isnan(c).any() | np.isnan(c.mean()):
            cadence = 0.0
        else:
            cadence = 1.0 / c.mean()
        # time_diff = sel[self.mjd_col][1:]-sel[self.mjd_col][:-1]
        r.append((field_ra, field_dec, band, np.mean(sel[self.m5_col]), cadence))

        res = np.rec.fromrecords(r, names=["field_ra", "field_dec", "band", "m5_mean", "cadence_mean"])

        zref = self.lim_sn.interp_griddata(res)

        if np.isnan(zref):
            zref = self.badval

        return zref



__all__ = ("StarCountMetric",)

import numpy as np

from rubin_sim.maf.metrics import BaseMetric

from .star_counts import *

# Example for CountMetric
# Mike Lund - Vanderbilt University
# mike.lund@gmail.com
# Last edited 8/15/2015
# Motivation: The distances to stars in LSST will be signficant enough that the structure of the Galaxy will be readily apparent because of its influence on the number of stars in a given field. Any metric concerned with the number of potential objects to be detected will need to feature not only the effects of the cadence but also the number of objects per field.
# This metric identifies the number of stars in a given field in a particular distance range. D1 and D2 are the close and far distances in parsecs.
# Requires StarCounts.StarCounts


class StarCountMetric(BaseMetric):
    """Find the number of stars in a given field between d1 and d2 in parsecs."""

    def __init__(self, **kwargs):
        self.d1 = kwargs.pop("D1", 100)
        self.d2 = kwargs.pop("D2", 1000)
        super(StarCountMetric, self).__init__(col=[], **kwargs)

    def run(self, data_slice, slice_point=None):
        self.dec_col = np.degrees(data_slice[0][3])
        self.ra_col = np.degrees(data_slice[0][2])
        return starcount.starcount(self.ra_col, self.dec_col, self.d1, self.d2)



__all__ = (
    "optimal_bins",
    "percentile_clipping",
    "radec2pix",
    "collapse_night",
    "load_inst_zeropoints",
)

import os
import warnings

import healpy as hp
import numpy as np
from rubin_scheduler.data import get_data_dir
from rubin_scheduler.utils import SysEngVals, int_binned_stat
from scipy.stats import binned_statistic

from rubin_sim.phot_utils import Bandpass, PhotometricParameters


def load_inst_zeropoints():
    """Load up and return instumental zeropoints and atmospheric extinctions"""
    zp_inst = {}
    datadir = get_data_dir()
    for filtername in "ugrizy":
        # set gain and exptime to 1 so the instrumental zeropoint will be in photoelectrons and per second
        phot_params = PhotometricParameters(nexp=1, gain=1, exptime=1, bandpass=filtername)
        bp = Bandpass()
        bp.read_throughput(os.path.join(datadir, "throughputs/baseline/", "total_%s.dat" % filtername))
        zp_inst[filtername] = bp.calc_zp_t(phot_params)

    syseng = SysEngVals()
    k_atm = syseng.k_atm

    return zp_inst, k_atm


def coadd_m5(mags):
    """Coadded depth, assuming Gaussian noise"""
    return 1.25 * np.log10(np.sum(10.0 ** (0.8 * np.array(mags))))


def collapse_night(
    data_slice,
    night_col="night",
    filter_col="filter",
    m5_col="fiveSigmaDepth",
    mjd_col="observationStartMJD",
):
    """Collapse a data_slice into per-filter, per-night values for
    the 'night', 'filter', 'median observationStartMJD', and 'fiveSigmaDepth'.
    """
    filters = np.unique(data_slice[filter_col])
    # Find the per-night, per-filter values
    night_slice = {}
    for filtername in filters:
        infilt = np.where(data_slice[filter_col] == filtername)[0]
        unight = np.unique(data_slice[night_col][infilt])
        right = unight + 0.5
        bins = [unight[0] - 0.5] + right.tolist()
        coadds, be, bn = binned_statistic(
            data_slice[night_col][infilt],
            data_slice[m5_col][infilt],
            bins=bins,
            statistic=coadd_m5,
        )

        unights, median_mjd_per_night = int_binned_stat(
            data_slice[night_col][infilt],
            data_slice[mjd_col][infilt],
            statistic=np.median,
        )

        night_slice[filtername] = np.array(
            list(zip(unight, median_mjd_per_night, coadds, filtername * len(unight))),
            dtype=[
                (night_col, int),
                (mjd_col, float),
                (m5_col, float),
                (filter_col, "U1"),
            ],
        )

    night_slice = np.concatenate([night_slice[f] for f in night_slice])
    night_slice.sort(order=["observationStartMJD"])

    return night_slice


def optimal_bins(datain, binmin=None, binmax=None, nbin_max=200, nbin_min=1):
    """
    Set an 'optimal' number of bins using the Freedman-Diaconis rule.

    Parameters
    ----------
    datain : numpy.ndarray or numpy.ma.MaskedArray
        The data for which we want to set the bin_size.
    binmin : float
        The minimum bin value to consider (if None, uses minimum data value).
    binmax : float
        The maximum bin value to consider (if None, uses maximum data value).
    nbin_max : int
        The maximum number of bins to create. Sometimes the 'optimal bin_size' implies
        an unreasonably large number of bins, if the data distribution is unusual.
    nbin_min : int
        The minimum number of bins to create. Default is 1.

    Returns
    -------
    int
        The number of bins.
    """
    # if it's a masked array, only use unmasked values
    if hasattr(datain, "compressed"):
        data = datain.compressed()
    else:
        data = datain
    # Check that any good data values remain.
    if data.size == 0:
        nbins = nbin_max
        warnings.warn(
            "No unmasked data available for calculating optimal bin size: returning %i bins" % (nbins)
        )
    # Else proceed.
    else:
        if binmin is None:
            binmin = np.nanmin(data)
        if binmax is None:
            binmax = np.nanmax(data)
        cond = np.where((data >= binmin) & (data <= binmax))[0]
        # Check if any data points remain within binmin/binmax.
        if np.size(data[cond]) == 0:
            nbins = nbin_max
            warnings.warn(
                "No data available for calculating optimal bin size within range of %f, %f" % (binmin, binmax)
                + ": returning %i bins" % (nbins)
            )
        else:
            iqr = np.percentile(data[cond], 75) - np.percentile(data[cond], 25)
            binwidth = 2 * iqr * (np.size(data[cond]) ** (-1.0 / 3.0))
            nbins = (binmax - binmin) / binwidth
            if nbins > nbin_max:
                warnings.warn(
                    "Optimal bin calculation tried to make %.0f bins, returning %i" % (nbins, nbin_max)
                )
                nbins = nbin_max
            if nbins < nbin_min:
                warnings.warn(
                    "Optimal bin calculation tried to make %.0f bins, returning %i" % (nbins, nbin_min)
                )
                nbins = nbin_min
    if np.isnan(nbins):
        warnings.warn("Optimal bin calculation calculated NaN: returning %i" % (nbin_max))
        nbins = nbin_max
    return int(nbins)


def percentile_clipping(data, percentile=95.0):
    """
    Calculate the minimum and maximum values of a distribution of points, after
    discarding data more than 'percentile' from the median.
    This is useful for determining useful data ranges for plots.
    Note that 'percentile' percent of the data is retained.

    Parameters
    ----------
    data : numpy.ndarray
        The data to clip.
    percentile : float
        Retain values within percentile of the median.

    Returns
    -------
    float, float
        The minimum and maximum values of the clipped data.
    """
    lower_percentile = (100 - percentile) / 2.0
    upper_percentile = 100 - lower_percentile
    min_value = np.percentile(data, lower_percentile)
    max_value = np.percentile(data, upper_percentile)
    return min_value, max_value


def radec2pix(nside, ra, dec):
    """
    Calculate the nearest healpixel ID of an RA/Dec array, assuming nside.

    Parameters
    ----------
    nside : int
        The nside value of the healpix grid.
    ra : numpy.ndarray
        The RA values to be converted to healpix ids, in radians.
    dec : numpy.ndarray
        The Dec values to be converted to healpix ids, in radians.

    Returns
    -------
    numpy.ndarray
        The healpix ids.
    """
    lat = np.pi / 2.0 - dec
    hpid = hp.ang2pix(nside, lat, ra)
    return hpid



#####################################################################################################
# Purpose: Calculate the galaxy counts for each Healpix pixel directly.
# Necessary when accounting for pixel-specific calibration errors (as they modify the magnitude limit
# to which incompleteness-corrected galaxy LF is integrated over).
#
# Similar to GalaxyCountsMetric_extended but does the analysis on each HEALpix pixel individually,
# without communicating with the slicer. Like a psuedo-metric. Accomodates individual redshift
# bins; galaxy LF powerlaws based on mock catalogs from Padilla et al.
#
# Need constantsForPipeline.py to import the power law constants and the normalization factor.
#
# Humna Awan: humna.awan@rutgers.edu
#####################################################################################################
__all__ = ("galaxy_counts_with_pixel_calibration",)

import warnings

import numpy as np
import scipy

from rubin_sim.maf.maf_contrib.lss_obs_strategy.constants_for_pipeline import (
    normalization_constant,
    power_law_const_a,
    power_law_const_b,
)


def galaxy_counts_with_pixel_calibration(
    coaddm5,
    upper_mag_limit,
    nside=128,
    filter_band="i",
    redshift_bin="all",
    cfhtls_counts=False,
    normalized_mock_catalog_counts=True,
):
    """

    Estimate galaxy counts for a given HEALpix pixel directly (without a slicer).

    Parameters
    ----------
    coaddm5 : `float`
        coadded 5sigma limiting magnitude for the pixel.
    upper_mag_limit : `float`
        upper limit on the magnitude, used to calculate num_gal.
    nside: `int`, opt
        HEALpix resolution parameter. Default: 128
    filter_band : `str`, opt
        Any one of 'u', 'g', 'r', 'i', 'z', 'y'. Default: 'i'
    redshift_bin : `str`, opt
        options include '0.<z<0.15', '0.15<z<0.37', '0.37<z<0.66, '0.66<z<1.0',
        '1.0<z<1.5', '1.5<z<2.0', '2.0<z<2.5', '2.5<z<3.0','3.0<z<3.5', '3.5<z<4.0',
        'all' for no redshift restriction (i.e. 0.<z<4.0)
        Default: 'all'
    cfhtls_counts : `bool`, opt
        set to True if want to calculate the total galaxy counts from CFHTLS
        powerlaw from LSST Science Book. Must be run with redshift_bin= 'all'
        Default: False
    normalized_mock_catalog_counts: `bool`, opt
        set to False if  want the raw/un-normalized galaxy counts from mock catalogs.
        Default: True

    """
    # Need to scale down to indivdual HEALpix pixels. Galaxy count from the Coadded depth is per 1 square degree.
    # Number of galaxies ~= 41253 sq. degrees in the full sky divided by number of HEALpix pixels.
    scale = 41253.0 / (int(12) * nside**2)
    # Reset units (otherwise uses magnitudes).
    units = "Galaxy Counts"

    # ------------------------------------------------------------------------
    # calculate the change in the power law constant based on the band
    # colors assumed here: (u-g)=(g-r)=(r-i)=(i-z)= (z-y)=0.4
    band_correction = -100
    if filter_band == "u":  # dimmer than i: u-g= 0.4 => g= u-0.4 => i= u-0.4*3
        band_correction = -0.4 * 3.0
    elif filter_band == "g":  # dimmer than i: g-r= 0.4 => r= g-0.4 => i= g-0.4*2
        band_correction = -0.4 * 2.0
    elif filter_band == "r":  # dimmer than i: i= r-0.4
        band_correction = -0.4
    elif filter_band == "i":  # i
        band_correction = 0.0
    elif filter_band == "z":  # brighter than i: i-z= 0.4 => i= z+0.4
        band_correction = 0.4
    elif filter_band == "y":  # brighter than i: z-y= 0.4 => z= y+0.4 => i= y+0.4*2
        band_correction = 0.4 * 2.0
    else:
        print("ERROR: Invalid band in GalaxyCountsMetric_withPixelCalibErrors. Assuming i-band.")
        band_correction = 0

    # ------------------------------------------------------------------------
    # check to make sure that the z-bin assigned is valid.
    if (redshift_bin != "all") and (redshift_bin not in list(power_law_const_a.keys())):
        print(
            "ERROR: Invalid redshift bin in GalaxyCountsMetric_withPixelCalibration. Defaulting to all redshifts."
        )
        redshift_bin = "all"

    # ------------------------------------------------------------------------
    # set up the functions for the integrand
    # when have a redshift slice
    def gal_count_bin(apparent_mag, coaddm5):
        dn_gal = np.power(
            10.0,
            power_law_const_a[redshift_bin] * (apparent_mag + band_correction)
            + power_law_const_b[redshift_bin],
        )
        completeness = 0.5 * scipy.special.erfc(apparent_mag - coaddm5)
        return dn_gal * completeness

    # when have to consider the entire z-range
    def gal_count_all(apparent_mag, coaddm5):
        if cfhtls_counts:
            # LSST power law: eq. 3.7 from LSST Science Book converted to per sq degree:
            # (46*3600)*10^(0.31(i-25))
            dn_gal = 46.0 * 3600.0 * np.power(10.0, 0.31 * (apparent_mag + band_correction - 25.0))
        else:
            # full z-range considered here: 0.<z<4.0
            # sum the galaxy counts from each individual z-bin
            dn_gal = 0.0
            for key in list(power_law_const_a.keys()):
                dn_gal += np.power(
                    10.0,
                    power_law_const_a[key] * (apparent_mag + band_correction) + power_law_const_b[key],
                )
        completeness = 0.5 * scipy.special.erfc(apparent_mag - coaddm5)
        return dn_gal * completeness

    # ------------------------------------------------------------------------
    # some coaddm5 values come out really small (i.e. min= 10**-314). Zero them out.
    if coaddm5 < 1:
        coaddm5 = 0

    # Calculate the number of galaxies.
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        # set up parameters to consider individual redshift range
        if redshift_bin == "all":
            num_gal, int_err = scipy.integrate.quad(gal_count_all, -np.inf, upper_mag_limit, args=coaddm5)
        else:
            num_gal, int_err = scipy.integrate.quad(gal_count_bin, -np.inf, upper_mag_limit, args=coaddm5)

    if normalized_mock_catalog_counts and not cfhtls_counts:
        # Normalize the counts from mock catalogs to match up to CFHTLS counts fori<25.5 galaxy catalog
        # Found the scaling factor separately.
        num_gal = normalization_constant * num_gal

    # coaddm5=0 implies no observation. Set no observation to zero to num_gal.
    if coaddm5 < 1.0:
        num_gal = 0.0
    if num_gal < 1.0:
        num_gal = 0.0

    # scale down to individual HEALpix pixel
    num_gal *= scale
    return num_gal
